import os

from pypads.functions.analysis.call_tracker import LoggingEnv
from pypads.functions.loggers.base_logger import LoggingFunction
from pypads.logging_util import WriteFormats, try_write_artifact


class Input(LoggingFunction):
    """
    Function logging the input parameters of the current pipeline object function call.
    """

    def __pre__(self, ctx, *args, _pypads_write_format=WriteFormats.pickle, _pypads_env: LoggingEnv, **kwargs):
        """
        :param ctx:
        :param args:
        :param _pypads_write_format:
        :param kwargs:
        :return:
        """
        for i in range(len(args)):
            arg = args[i]
            name = os.path.join(_pypads_env.call.to_folder(),
                                "args",
                                str(i) + "_" + str(id(_pypads_env.callback)))
            try_write_artifact(name, arg, _pypads_write_format)

        for (k, v) in kwargs.items():
            name = os.path.join(_pypads_env.call.to_folder(),
                                "kwargs",
                                str(k) + "_" + str(id(_pypads_env.callback)))
            try_write_artifact(name, v, _pypads_write_format)


class Output(LoggingFunction):
    """
    Function logging the output of the current pipeline object function call.
    """

    def __post__(self, ctx, *args, _pypads_write_format=WriteFormats.pickle, _pypads_env, **kwargs):
        """
        :param ctx:
        :param args:
        :param _pypads_write_format:
        :param kwargs:
        :return:
        """
        name = os.path.join(_pypads_env.call.to_folder(),
                            "returns",
                            str(id(_pypads_env.callback)))
        try_write_artifact(name, kwargs["_pypads_result"], _pypads_write_format)



import os

import mlflow
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger
from pypads.functions.analysis.call_tracker import LoggingEnv
from pypads.functions.loggers.base_logger import LoggingFunction
from pypads.logging_util import WriteFormats, get_base_folder, try_write_artifact
from pypads.util import is_package_available


# last_pipeline_tracking = None
#
# _pipeline_type = None
# network = None


# --- Clean nodes after run ---
def end_run(pads, *args, **kwargs):
    # curr_call = pads.call_tracker.current_call()
    pipeline_cache = pads.cache.get("pipeline", {})

    network = pipeline_cache.get("network", None)
    _pipeline_type = pipeline_cache.get("pipeline_type")
    # global network
    if network is not None and len(network.nodes) > 0:
        from networkx import DiGraph
        from networkx.drawing.nx_agraph import to_agraph
        try_write_artifact("_pypads_pipeline", network, WriteFormats.pickle)

        if is_package_available("networkx"):
            base_folder = get_base_folder()
            folder = base_folder + "pipeline_graph.png"
            if not os.path.exists(base_folder):
                os.mkdir(base_folder)
            if is_package_available("agraph") and is_package_available("graphviz"):
                try:
                    if _pipeline_type == "simple":
                        agraph = to_agraph(DiGraph(network))
                    elif _pipeline_type == "grouped":
                        copy = network.copy()
                        edge_groups = {}
                        for edge in copy.edges:
                            plain = copy.get_edge_data(*edge)['plain_label']
                            if plain not in edge_groups:
                                edge_groups[plain] = []
                            edge_groups[plain].append(edge)

                        for group, edges in edge_groups.items():
                            label = ""
                            suffix = ""
                            base_edge = None
                            for edge in edges:
                                splits = copy.get_edge_data(*edge)['label'].split(":")
                                if label == "":
                                    label = splits[0]
                                    suffix = splits[1]
                                    base_edge = edge
                                else:
                                    label = label + ", " + splits[0]
                                copy.remove_edge(*edge)
                            label = label + ":" + suffix
                            copy.add_edge(*base_edge, label=label)
                        agraph = to_agraph(copy)
                    elif _pipeline_type == "grouped_no_count":
                        copy = network.copy()
                        edge_groups = {}
                        for edge in copy.edges:
                            plain = copy.get_edge_data(*edge)['plain_label']
                            if plain not in edge_groups:
                                edge_groups[plain] = edge

                        copy.remove_edges_from(network.edges())
                        for group, edge in edge_groups.items():
                            copy.add_edge(*edge, label=group)
                        agraph = to_agraph(copy)
                    else:
                        agraph = to_agraph(network)
                    agraph.layout('dot')
                    agraph.draw(folder)
                except ValueError as e:
                    logger.warning("Failed plotting pipeline: " + str(e))
            elif is_package_available("matplotlib"):
                import matplotlib.pyplot as plt
                import networkx as nx
                pos = nx.spring_layout(network)
                nx.draw(network, pos)
                nx.draw_networkx_labels(network, pos=pos)
                nx.draw_networkx_edge_labels(network, pos)
                plt.savefig(folder)
            if os.path.exists(folder):
                try_mlflow_log(mlflow.log_artifact, folder)

    pads.cache.pop("pipeline")
    # # global last_pipeline_tracking
    # # last_pipeline_tracking = None
    # pads.cache.pop(curr_call)


# !--- Clean nodes after run ---


def _to_node_id(wrappe, ref):
    if ref is not None:
        return id(ref)
    else:
        return id(wrappe)


def _to_node_label(wrappe, ref):
    if ref is not None:
        try:
            return str(ref)
        except Exception as e:
            try:
                logger.warning(
                    "Couldn't get str representation for given ref of type " + str(
                        type(ref)) + ". Falling back to " + str(
                        wrappe) + " and id or ref " + str(id(ref)) + ". " + str(e))
                return str(wrappe) + str(id(ref))
            except Exception as e:
                logger.warning("Couldn't get fallback string. Fallback to id " + str(id(ref)) + ". " + str(e))
                return str(id(ref))
    else:
        return str(wrappe)


def _to_edge_label(wrappe, include_args, args, kwargs):
    if include_args:
        return str(wrappe) + " args: " + str(args) + " kwargs: " + str(kwargs)
    return str(wrappe)


def _step_number(network, label):
    return str(network.number_of_edges()) + ": " + label


class PipelineTracker(LoggingFunction):

    def _needed_packages(self):
        return ["networkx"]

    def __pre__(self, ctx, *args, _pypads_env: LoggingEnv, _pypads_pipeline_type="normal", _pypads_pipeline_args=False,
                **kwargs):

        from pypads.pypads import get_current_pads
        pads = get_current_pads()
        pads.api.register_post_fn("pipeline_clean_up", end_run)


        if pads.cache.exists("pipeline"):
            pipeline_cache = pads.cache.get("pipeline")
        else:
            pipeline_cache = {"network": None, "last_pipeline_tracking":None, "pipeline_type":_pypads_pipeline_type}

        network = pipeline_cache.get("network")
        last_pipeline_tracking = pipeline_cache.get("last_pipeline_tracking")
        _pipeline_type = pipeline_cache.get("pipeline_type")

        import networkx as nx
        if network is None:
            network = nx.MultiDiGraph()

        node_id = _to_node_id(_pypads_env.call.call_id.wrappee, ctx)
        label = _to_node_label(_pypads_env.call.call_id.wrappee, ctx)
        if not network.has_node(node_id):
            network.add_node(node_id, label=label)

        label = _to_edge_label(_pypads_env.call.call_id.wrappee, _pypads_pipeline_args, args, kwargs)
        # If the current stack holds only the call itself
        if pads.call_tracker.call_depth() == 1:

            # If there where no nodes until now
            if len(network.nodes) == 1:
                network.add_node(-1, label="entry")
                last_pipeline_tracking = -1
                pipeline_cache["last_pipeline_tracking"] = -1
            network.add_edge(last_pipeline_tracking, node_id, plain_label=label, label=_step_number(network,label))

        # If the tracked function was called from another tracked function
        elif pads.call_tracker.call_depth() > 1:

            containing_node_id = _to_node_id(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
            if not network.has_node(containing_node_id):
                containing_node_label = _to_node_label(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
                network.add_node(containing_node_id, label=containing_node_label)
            # Add an edge from the tracked function to the current function call
            network.add_edge(containing_node_id, node_id, plain_label=label, label=_step_number(network,label))

        pipeline_cache["network"] = network
        pads.cache.add("pipeline",pipeline_cache)
        return node_id

    def __post__(self, ctx, *args, _pypads_pipeline_args=False, _pypads_env: LoggingEnv, _pypads_pre_return, **kwargs):
        from pypads.pypads import get_current_pads
        pads = get_current_pads()

        pipeline_cache = pads.cache.get("pipeline")
        network = pipeline_cache.get("network")
        node_id = _pypads_pre_return
        label = "return " + _to_edge_label(_pypads_env.call.call_id.wrappee, _pypads_pipeline_args, args, kwargs)
        if pads.call_tracker.call_depth() == 1:
            network.add_edge(node_id, -1, plain_label=label, label=_step_number(network,label))
        elif pads.call_tracker.call_depth() > 1:
            containing_node_id = _to_node_id(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
            if not network.has_node(containing_node_id):
                containing_node_label = _to_node_label(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
                network.add_node(containing_node_id, label=containing_node_label)
            network.add_edge(node_id, containing_node_id, plain_label=label, label=_step_number(network,label))
        pipeline_cache["network"] = network
        pads.cache.add("pipeline", pipeline_cache)



from _py_abc import ABCMeta
from abc import abstractmethod

from pypads.functions.analysis.time_keeper import timed
from pypads.util import is_package_available


class NoCallAllowedError(Exception):
    """
    Exception to denote that a callable couldn't be called, but isn't essential.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MissingDependencyError(NoCallAllowedError):
    """
    Exception to be thrown if a dependency is missing.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SuperStop:
    """
    This class resolves the issue TypeError: object.__init__() takes exactly one argument by being the last class
    in a mro and ommitting all arguments. This should be always last in the mro()!
    """

    def __init__(self, *args, **kwargs):
        mro = self.__class__.mro()
        if SuperStop in mro:
            if len(mro) - 2 != mro.index(SuperStop):
                raise ValueError("SuperStop ommitting arguments in " + str(self.__class__)
                                 + " super() callstack: " + str(mro))
        super().__init__()


class OrderMixin(SuperStop):
    """
    Object defining an order attribute to denote its priority.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, order=0, **kwargs):
        self._order = order
        super().__init__(*args, **kwargs)

    def order(self):
        return self._order


class CallableMixin(SuperStop):
    """
    Object defining a _call method which can be overwritten.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.__real_call__(*args, **kwargs)

    @abstractmethod
    def __real_call__(self, *args, **kwargs):
        pass


class DependencyMixin(CallableMixin):
    """
    Callable being able to be disabled / enabled depending on package availability.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def _needed_packages():
        """
        :return: List of needed packages by the logger.
        """
        return []

    def _check_dependencies(self):
        """
        Raise error if dependencies are missing.

        """
        missing = []
        for package in self._needed_packages():
            if not is_package_available(package):
                missing.append(package)
        if len(missing) > 0:
            raise MissingDependencyError("Can't log " + str(self) + ". Missing dependencies: " + ", ".join(missing))

    def __call__(self, *args, **kwargs):
        self._check_dependencies()
        return super().__call__(*args, **kwargs)


class IntermediateCallableMixin(CallableMixin):
    """
    Callable being able to be disable / enabled on nested / intermediate runs.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, nested=True, intermediate=True, **kwargs):
        self._intermediate = intermediate
        self._nested = nested
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        from pypads.pypads import is_nested_run
        if self._nested or not is_nested_run():
            from pypads.pypads import is_intermediate_run
            if self._intermediate or not is_intermediate_run():
                return super().__call__(*args, **kwargs)
        raise NoCallAllowedError("Call wasn't allowed by intermediate / nested settings of the current run.")

    @property
    def nested(self):
        return self._nested

    @property
    def intermediate(self):
        return self._intermediate


class TimedCallableMixin(CallableMixin):
    __metaclass__ = ABCMeta
    """
    Callable tracking its own execution time.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        c = super().__call__
        _return, time = timed(lambda: c(*args, **kwargs))
        return _return, time


class DefensiveCallableMixin(CallableMixin):
    __metaclass__ = ABCMeta
    """
    Callable handling errors produced by itself.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, ctx, *args, _pypads_env=None, **kwargs):
        try:
            return super().__call__(ctx, *args, _pypads_env=_pypads_env, **kwargs)
        except Exception as e:
            return self._handle_error(*args, ctx=ctx, _pypads_env=_pypads_env, error=e, **kwargs)

    @abstractmethod
    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        raise NotImplementedError()


class ConfigurableCallableMixin(CallableMixin):
    __metaclass__ = ABCMeta
    """
    Callable storing additional creation args as fields to be accessible later on.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        self._kwargs = kwargs
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        super().__call__(*args, **{**self._kwargs, **kwargs})



import io
import pickle


def _pickle_tuple(*args):
    with io.BytesIO() as file:
        pickle.dump(tuple(args), file)
        file.seek(0)
        return file.read()


def _cloudpickle_tuple(*args):
    from joblib.externals.cloudpickle import dumps
    return dumps(tuple(args))



import os
import time
from functools import wraps
from typing import List

from pypads.util import is_package_available

if is_package_available("joblib"):
    import joblib

    original_delayed = joblib.delayed

    @wraps(original_delayed)
    def punched_delayed(fn):
        """Decorator used to capture the arguments of a function."""

        @wraps(fn)
        def wrapped_function(*args, _pypads_cache=None, _pypads_config=None, _pypads_active_run_id=None,
                             _pypads_tracking_uri=None,
                             _pypads_affected_modules=None, _pypads_triggering_process=None, **kwargs):
            from pypads.parallel.util import _pickle_tuple, _cloudpickle_tuple
            from pypads import logger

            # only if pads data was passed
            if _pypads_active_run_id:
                # noinspection PyUnresolvedReferences
                import pypads.pypads
                import mlflow

                is_new_process = not pypads.pypads.current_pads

                # If pads has to be reinitialized
                if is_new_process:
                    import pypads

                    # reactivate this run in the foreign process
                    mlflow.set_tracking_uri(_pypads_tracking_uri)
                    mlflow.start_run(run_id=_pypads_active_run_id, nested=True)

                    start_time = time.time()
                    logger.debug("Init Pypads in:" + str(time.time() - start_time))

                    from pypads.base import PyPads
                    _pypads = PyPads(uri=_pypads_tracking_uri, reload_warnings=False,
                                     config=_pypads_config,
                                     affected_modules=_pypads_affected_modules,
                                     clear_imports=True, pre_initialized_cache=_pypads_cache, reload_modules=True,
                                     disable_run_init=True)

                    def clear_mlflow():
                        """
                        Don't close run. This function clears the run which was reactivated from the stack to stop a closing of it.
                        :return:
                        """
                        if len(mlflow.tracking.fluent._active_run_stack) == 1:
                            mlflow.tracking.fluent._active_run_stack.pop()

                    import atexit
                    atexit.register(clear_mlflow)

                # If pads already exists on process
                else:
                    _pypads = pypads.pypads.current_pads
                    _pypads.cache.merge(_pypads_cache)

                # Unpickle args
                from pickle import loads
                start_time = time.time()
                a, b = loads(args[0])
                logger.debug("Loading args from pickle in:" + str(time.time() - start_time))

                # Unpickle function
                from cloudpickle import loads as c_loads
                start_time = time.time()
                wrapped_fn = c_loads(args[1])[0]
                logger.debug("Loading punched function from pickle in:" + str(time.time() - start_time))

                args = a
                kwargs = b

                logger.debug("Started wrapped function on process: " + str(os.getpid()))

                out = wrapped_fn(*args, **kwargs)
                return out, _pypads.cache

            else:
                return fn(*args, **kwargs)

        def delayed_function(*args, **kwargs):
            from pypads.parallel.util import _pickle_tuple, _cloudpickle_tuple
            import mlflow
            run = mlflow.active_run()
            if run:
                from pypads.pypads import current_pads
                if current_pads and current_pads.config["track_sub_processes"]:
                    # TODO Only cloudpickle args / kwargs if needed and not always.
                    pickled_params = (_pickle_tuple(args, kwargs), _cloudpickle_tuple(fn))
                    args = pickled_params
                    from pypads.pypads import get_current_pads

                    pads = get_current_pads()

                    # TODO Pickle all for reinitialisation important things (Logging functions, config, init run fns)
                    kwargs = {"_pypads_cache": pads.cache,
                              "_pypads_config": pads.config,
                              "_pypads_active_run_id": run.info.run_id,
                              "_pypads_tracking_uri": pads.tracking_uri,
                              "_pypads_affected_modules": pads.wrap_manager.module_wrapper.punched_module_names,
                              "_pypads_triggering_process": os.getpid()}
            from pypads.pads_loguru import logger_manager
            logger_manager.temporary_remove()
            return wrapped_function, args, kwargs

        try:
            import functools
            delayed_function = functools.wraps(fn)(delayed_function)
        except AttributeError:
            " functools.wraps fails on some callable objects "
        return delayed_function


    setattr(joblib, "delayed", punched_delayed)

    # original_dispatch = joblib.Parallel._dispatch
    #
    # def _dispatch(self, *args, **kwargs):
    #     print(self._backend)
    #     out = original_dispatch(self, *args, **kwargs)
    #     return out
    #
    # joblib.Parallel._dispatch = _dispatch

    original_call = joblib.Parallel.__call__


    @wraps(original_call)
    def joblib_call(self, *args, **kwargs):
        from pypads.caches import PypadsCache
        from pypads import logger
        from pypads.pypads import current_pads
        pads = current_pads

        if pads:
            if pads.config["track_sub_processes"]:
                # Temporary hold handlers and remove them
                from pypads.pads_loguru import logger_manager
                logger_manager.temporary_remove()
                out = original_call(self, *args, **kwargs)
                if isinstance(out, List):
                    real_out = []
                    for entry in out:
                        if isinstance(entry, tuple) and len(entry) == 2 and isinstance(entry[1], PypadsCache):
                            real_out.append(entry[0])
                            cache = entry[1]
                            pads.cache.merge(cache)
                        else:
                            real_out.append(entry)
                    out = real_out
                logger_manager.add_loggers_from_history()
                return out
            else:
                logger.warning(
                    "Call of joblib parallel found with self: " + str(self) + " args: " + str(args) + "kwargs: " + str(
                        kwargs) + " but subprocess tracking is deactivated. To activated subprocess tracking set "
                                  "config parameter track_sub_processes to true. Disclaimer: this might be currently "
                                  "unstable and/or bad for the performance.")

        from pypads.pads_loguru import logger_manager
        logger_manager.temporary_remove()
        out = original_call(self, *args, **kwargs)
        logger_manager.add_loggers_from_history()
        return out


    setattr(joblib.Parallel, "__call__", joblib_call)



import glob
import json
import os
from itertools import chain
from os.path import expanduser
from typing import List

from pypads import logger
from pypads.autolog.hook import get_hooks

mapping_files = glob.glob(expanduser("~") + ".pypads/bindings/**.json")
mapping_files.extend(
    glob.glob(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/bindings/resources/mapping/**.json"))


class AlgorithmMeta:
    def __init__(self, name, concepts):
        self._concepts = concepts
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def concepts(self):
        return self._concepts

    def __eq__(self, other):
        return self.name == other.name


class AlgorithmMapping:
    """
    Mapping for an algorithm defined by a pypads mapping file
    """

    def __init__(self, reference, library, algorithm: AlgorithmMeta, file, hooks):
        self._hooks = hooks
        self._algorithm = algorithm
        self._library = library
        self._reference = reference
        self._file = file
        self._in_collection = None

    @property
    def in_collection(self):
        return self._in_collection

    @in_collection.setter
    def in_collection(self, value):
        self._in_collection = value

    @property
    def file(self):
        return self._file

    @property
    def reference(self):
        return self._reference

    @property
    def library(self):
        return self._library

    @property
    def algorithm(self):
        return self._algorithm

    @property
    def hooks(self):
        return self._hooks

    @hooks.setter
    def hooks(self, value):
        if not isinstance(value, List):
            self._hooks = get_hooks(value)
        else:
            self._hooks = value

    def __str__(self):
        return "Mapping[" + str(self.file) + ":" + str(self.reference) + ", lib=" + str(self.library) + ", alg=" + str(
            self.algorithm) + ", hooks=" + str(self.hooks) + "]"

    def __eq__(self, other):
        # TODO also check for reference, file, in_collection?
        if self.hooks and other.hooks:
            return set(self.hooks) == set(
                other.hooks) and self.algorithm == other.algorithm and self.library == other.library
        elif not self.hooks and not other.hooks:
            return self.algorithm == other.algorithm and self.library == other.library
        else:
            return False


class MappingCollection:
    def __init__(self, key, default_hooks, algorithms: List[AlgorithmMapping]):
        self._default_hooks = default_hooks
        self._algorithms = algorithms
        self._key = key

    @property
    def key(self):
        return self._key

    @property
    def default_hooks(self):
        return self._default_hooks

    @property
    def algorithms(self):
        for alg in self._algorithms:
            hooks = None
            if "hooks" in alg:
                hooks = get_hooks(alg["hooks"])

            if alg["implementation"] and len(alg["implementation"]) > 0:
                for library, reference in alg["implementation"].items():
                    mapping = AlgorithmMapping(reference, library, AlgorithmMeta(alg, []), self.key, hooks)
                    mapping.in_collection = self
                    yield mapping

    def get_default_module_hooks(self):
        if "modules" in self._default_hooks:
            if "fns" in self._default_hooks["modules"]:
                return get_hooks(self._default_hooks["modules"]["fns"])

    def get_default_class_hooks(self):
        if "classes" in self._default_hooks:
            if "fns" in self._default_hooks["classes"]:
                return get_hooks(self._default_hooks["classes"]["fns"])

    def get_default_fn_hooks(self):
        if "fns" in self._default_hooks:
            return get_hooks(self._default_hooks["fns"])


class MappingFile(MappingCollection):

    def __init__(self, name, json):
        super().__init__(name, json['default_hooks'] if 'default_hooks' in json else {
            "modules": {
                "fns": {}
            },
            "classes": {
                "fns": {}
            },
            "fns": {}
        }, json["algorithms"] if 'algorithms' in json else [])
        self._lib = json['metadata']['library']
        self._lib_version = json['metadata']['library_version']
        self._version = json['metadata']['mapping_version']
        self._name = name

    @property
    def version(self):
        return self._version

    @property
    def lib(self):
        return self._lib

    @property
    def lib_version(self):
        return self._lib_version

    @property
    def name(self):
        return self._name


class MappingRegistry:
    """
    Class holding all the mappings
    """

    def __init__(self, *paths):

        self._mappings = {}
        self.found_classes = {}

        for path in paths:
            self.load_mapping(path)

    def add_mapping(self, mapping: MappingCollection, key=None):
        if isinstance(mapping, MappingFile):
            key = mapping.lib

        if key is None:
            logger.error(
                "Couldn't add mapping " + str(mapping) + " to the pypads mapping registry. Lib or key are undefined.")
        else:
            self._mappings[key] = mapping

    def load_mapping(self, path):
        with open(path) as json_file:
            name = os.path.basename(json_file.name)
            logger.info("Added mapping file with name: " + str(name))
            content = json.load(json_file)
            self.add_mapping(MappingFile(name, json=content))

    def add_found_class(self, mapping):
        if mapping.reference not in self.found_classes:
            self.found_classes[mapping.reference] = mapping

    def iter_found_classes(self):
        for i, mapping in self.found_classes.items():
            yield mapping

    def get_libraries(self):
        all_libs = set()
        for key, mapping in self._mappings.items():
            all_libs.add(mapping.lib)
        return all_libs

    def get_relevant_mappings(self):
        """
        Function to find all relevant mappings. This produces a generator getting extended with found subclasses
        :return:
        """
        return chain(self.get_algorithms(), self.iter_found_classes())

    def get_algorithms(self):
        """
        Get all mappings defined in all mapping files.
        :return:
        """

        for key, mapping in self._mappings.items():
            for alg in mapping.algorithms:
                yield alg



import sys

import gorilla
from mlflow.utils import experimental

from pypads.functions.analysis.call_tracker import LoggingEnv
from pypads.functions.loggers.base_logger import LoggingFunction
from pypads.util import is_package_available

added_autologs = set()
mlflow_autolog_fns = {}
mlflow_autolog_callbacks = []


def _to_patch_id(patch):
    return str(patch.desitination)


def fake_gorilla_apply(patch):
    if patch.name not in mlflow_autolog_fns:
        mlflow_autolog_fns[patch.name] = {}
    mlflow_autolog_fns[patch.name][patch.destination] = patch


# For now only take last added callback
def fake_gorilla_get_original_attribute(clz, fn_name):
    return mlflow_autolog_callbacks.pop()


gorilla.apply = fake_gorilla_apply
gorilla.get_original_attribute = fake_gorilla_get_original_attribute


# Could also be a normal function right now
class MlflowAutologger(LoggingFunction):
    """
    MlflowAutologger is the intergration of the mlflow autologging functionalities into PyPads tracking system.

    """
    def __init__(self, *args, order=-1, **kwargs):
        super().__init__(*args, order=order, **kwargs)

    @experimental
    def __call_wrapped__(self, ctx, *args, _args, _kwargs, _pypads_autologgers=None, _pypads_env=LoggingEnv, **kwargs):
        """
            Function used to enable autologgers of mlflow.
        """

        if _pypads_autologgers is None:
            _pypads_autologgers = ["keras", "tensorflow", "xgboost", "gluon", "spark", "lightgbm"]

        if 'tensorflow' in _pypads_autologgers and 'tensorflow' in sys.modules and 'tensorflow' not in added_autologs and is_package_available(
                'tensorflow'):
            added_autologs.add('tensorflow')
            from mlflow import tensorflow
            tensorflow.autolog()

        if 'keras' in _pypads_autologgers and 'keras' in sys.modules and 'keras' not in added_autologs and is_package_available(
                'keras'):
            added_autologs.add('keras')
            from mlflow import keras
            keras.autolog()

        if 'xgboost' in _pypads_autologgers and 'xgboost' in sys.modules and 'xgboost' not in added_autologs and is_package_available(
                'xgboost'):
            added_autologs.add('xgboost')
            from mlflow import xgboost
            xgboost.autolog()

        if 'gluon' in _pypads_autologgers and 'gluon' in sys.modules and 'gluon' not in added_autologs and is_package_available(
                'gluon'):
            added_autologs.add('gluon')
            from mlflow import gluon
            gluon.autolog()

        if 'spark' in _pypads_autologgers and 'spark' in sys.modules and 'spark' not in added_autologs and is_package_available(
                'pyspark'):
            added_autologs.add('spark')
            from mlflow import spark
            spark.autolog()

        if 'lightgbm' in _pypads_autologgers and 'lightgbm' in sys.modules and 'lightgbm' not in added_autologs and is_package_available(
                'lightgbm'):
            added_autologs.add('lightgbm')
            from mlflow import lightgbm
            lightgbm.autolog()

        # If the function is to be logged call the related mlflow autolog function which would have
        #  been applied via gorilla
        if _pypads_env.call.call_id.wrappee.__name__ in mlflow_autolog_fns:
            for destination, patch in mlflow_autolog_fns[_pypads_env.call.call_id.wrappee.__name__].items():
                if destination == _pypads_env.call.call_id.context.container or issubclass(
                        _pypads_env.call.call_id.context.container, destination):
                    from pypads.autolog.wrapping.base_wrapper import Context
                    # Jump directly to the original function
                    mlflow_autolog_callbacks.append(Context(destination).original(getattr(destination, patch.name)))
                    return patch.obj(ctx, *_args, **_kwargs)
        return _pypads_env.callback(*_args, **_kwargs)

# # Override mlflow because of: Using functools.wraps unfortunately breaks with inspect.getargspec(fn) of mlflow
# def log_fn_args_as_params(fn, args, kwargs, unlogged=[]):  # pylint: disable=W0102
#     """
#     Log parameters explicitly passed to a function.
#     :param fn: function whose parameters are to be logged
#     :param args: arguments explicitly passed into fn
#     :param kwargs: kwargs explicitly passed into fn
#     :param unlogged: parameters not to be logged
#     :return: None
#     """
#     # all_default_values has length n, corresponding to values of the
#     # last n elements in all_param_names
#     import inspect
#     from mlflow.utils.autologging_utils import try_mlflow_log
#     all_param_names, _, _, all_default_values, _, _, _ = inspect.getfullargspec(fn)  # pylint: disable=W1505
#
#     # Checking if default values are present for logging. Known bug that getargspec will return an
#     # empty argspec for certain functions, despite the functions having an argspec.
#     if all_default_values is not None and len(all_default_values) > 0:
#         # Logging the default arguments not passed by the user
#         from mlflow.utils.autologging_utils import get_unspecified_default_args
#         defaults = get_unspecified_default_args(args, kwargs, all_param_names, all_default_values)
#
#         for name in [name for name in defaults.keys() if name in unlogged]:
#             del defaults[name]
#         try_mlflow_log(mlflow.log_params, defaults)
#
#     # Logging the arguments passed by the user
#     args_dict = dict((param_name, param_val) for param_name, param_val
#                      in zip(all_param_names[:len(args)], args)
#                      if param_name not in unlogged)
#     if len(args_dict.keys()) > 0:
#         try_mlflow_log(mlflow.log_params, args_dict)
#
#     # Logging the kwargs passed by the user
#     for param_name in kwargs:
#         if param_name not in unlogged:
#             try_mlflow_log(mlflow.log_param, param_name, kwargs[param_name])
#
#
# from mlflow.utils import autologging_utils
# autologging_utils.log_fn_args_as_params = log_fn_args_as_params



import mlflow
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger
from pypads.functions.loggers.base_logger import LoggingFunction
from pypads.logging_util import try_write_artifact, WriteFormats


class Metric(LoggingFunction):
    """
    Function logging the wrapped metric function
    """

    def __post__(self, ctx, *args, _pypads_artifact_fallback=False, _pypads_env, _pypads_result, **kwargs):
        """

        :param ctx:
        :param args:
        :param _pypads_artifact_fallback: Write to artifact if metric can not be logged as an double value into mlflow
        :param _pypads_result:
        :param kwargs:
        :return:
        """
        result = _pypads_result

        if result is not None:
            if isinstance(result, float):
                try_mlflow_log(mlflow.log_metric,
                               _pypads_env.call.call_id.context.container.__name__ + "." + _pypads_env.call.call_id.wrappee.__name__ + ".txt",
                               result, step=_pypads_env.call.call_id.call_number)
            else:
                logger.warning("Mlflow metrics have to be doubles. Could log the return value of type '" + str(
                    type(
                        result)) + "' of '" + _pypads_env.call.call_id.context.container.__name__ + "." + _pypads_env.call.call_id.wrappee.__name__ + "' as artifact instead.")

                # TODO search callstack for already logged functions and ignore?
                if _pypads_artifact_fallback:
                    logger.info(
                        "Logging result if '" + _pypads_env.call.call_id.context.container.__name__ + "' as artifact.")
                    try_write_artifact(_pypads_env.call.call_id.context.container.__name__, str(result),
                                       WriteFormats.text)



from pypads.functions.pre_run.pre_run import PreRunFunction
from pypads.functions.util import get_git_repo


class IGit(PreRunFunction):

    @staticmethod
    def _needed_packages():
        return ['git']

    def _call(self, pads, *args, **kwargs):
        _pypads_timeout = kwargs.get("_pypads_timeout") if kwargs.get("_pypads_timeout") else 5
        run = pads.api.active_run()
        tags = run.data.tags
        source_name = tags.get("mlflow.source.name", None)
        repo = get_git_repo(source_name, pads=pads) if source_name else None
        if repo:
            # Disable pager for returns
            repo.git.set_persistent_git_options(no_pager=True)
            pads.api.set_tag("pypads.git.description", repo.description)
            pads.api.set_tag("pypads.git.describe", repo.git.describe("--all"))
            from git import GitCommandError
            # try:
            #     pads.api.set_tag("pypads.git.shortlog", repo.git.shortlog(kill_after_timeout=_pypads_timeout))
            # except GitCommandError as e:
            #     warning("Ignored the execution and tracking of 'git shortlog'. " + str(e))
            try:
                pads.api.set_tag("pypads.git.log", repo.git.log(kill_after_timeout=_pypads_timeout))
            except GitCommandError as e:
                pass
            remotes = repo.remotes
            remote_out = "No remotes existing"
            if len(remotes) > 0:
                remote_out = ""
                for remote in remotes:
                    remote_out += remote.name + ": " + remote.url + "\n"
            pads.api.set_tag("pypads.git.remotes", remote_out)

#
# class GitMirror(PreRunFunction):
#
#     @staticmethod
#     def _needed_packages():
#         return ['git']
#
#     def _call(self, pads, *args, **kwargs):
#         if pads.config["mirror_git"]:
#             repo = get_run_git()
#             if repo:
#                 # mirror the given repo to our git remote server
#                 pass



import ast

import mlflow

from pypads import logger
from pypads.base import PyPads, CONFIG_NAME

# Cache configs for runs. Each run could is for now static in it's config.
configs = {}
current_pads = None

# !--- Clean the config cache after run ---


def set_current_pads(pads: PyPads):
    global current_pads
    current_pads = pads


def get_current_pads() -> PyPads:
    """
    Get the currently active pypads instance. All duck punched objects use this function for interacting with pypads.
    :return:
    """
    global current_pads
    if not current_pads:
        # Try to reload pads if it was already defined in the active run
        config = get_current_config()

        if config:
            logger.warning(
                "PyPads seems to be missing on given run with saved configuration. Reinitializing.")
            return PyPads(config=config)
        else:
            logger.warning(
                "PyPads has to be initialized before logging can be used. Initializing for your with default values.")
            return PyPads()
    return current_pads


def is_nested_run():
    pads = get_current_pads()
    tags = pads.mlf.get_run(pads.api.active_run().info.run_id).data.tags
    return "mlflow.parentRunId" in tags


def is_intermediate_run():
    pads = get_current_pads()
    return pads.api.is_intermediate_run()


def get_current_config(default=None):
    """
    Get configuration defined in the current mlflow run
    :return:
    """
    global configs
    active_run = mlflow.active_run()
    if active_run in configs.keys():
        return configs[active_run]
    if not active_run:
        return default
    run = mlflow.get_run(active_run.info.run_id)
    if CONFIG_NAME in run.data.tags:
        configs[active_run] = ast.literal_eval(run.data.tags[CONFIG_NAME])
        return configs[active_run]
    return default



import os

from pypads import logger
from pypads.util import is_package_available


def get_run_git():
    from pypads.pypads import get_current_pads
    from pypads.base import PyPads
    pads: PyPads = get_current_pads()
    run = pads.api.active_run()
    tags = run.data.tags
    source_name = tags.get("mlflow.source.name", None)
    if source_name:
        return get_git_repo(source_name)
    return None


def get_git_repo(path, pads=None):
    if is_package_available("git"):
        import git
        if os.path.isfile(path):
            path = os.path.dirname(path)
            from git import InvalidGitRepositoryError
            try:
                repo = git.Repo(path, search_parent_directories=True)
                check_index(repo, message="PyPads issued commit", pads=pads)
                return repo
            except InvalidGitRepositoryError:
                logger.warning("No existing git repository was found, initializing a new one...")
                return init_git_repo(path, pads=pads)
    else:
        logger.warning("Git needs to be installed to manage a central repository.")
        return None


def init_git_repo(path, pads=None):
    if is_package_available("git"):
        import git
        from git import InvalidGitRepositoryError, GitCommandError, GitError
        try:
            repo = git.Repo.init(path, bare=False)
            check_index(repo, message="PyPads initialized repository.", init=True, pads=pads)
            logger.info("Repository was successfully initialized")
            return repo
        except (InvalidGitRepositoryError, GitCommandError, GitError):
            logger.warning("Git could not initialize a repository in this directory %s" % path)
            return None
    else:
        logger.warning("Git needs to be installed to manage a central repository.")
        return None


def check_index(repo, message="", init=False, pads=None):
    if not pads:
        from pypads.base import get_current_pads, PyPads
        pads = get_current_pads()
    run = pads.api.active_run()
    if init:
        try:
            add_files(repo)
            repo.git.commit(message=message)
            pads.api.set_tag("pypads.source.git.commit", repo.head.object.hexsha)
            pads.api.set_tag("pypads.git.branch", repo.active_branch.name)

        except Exception as e:
            logger.warning("Could not issue the first commit due to this error %s" % str(e))
    else:
        try:
            orig_branch = repo.active_branch.name
            if len(repo.index.diff('HEAD')) > 0 or len(repo.index.diff(None)) > 0 or repo.untracked_files:
                logger.warning("There are uncommitted changes in your git!")
                # Save those changes to stash
                repo.git.stash('push', '--include-untracked')

                # check if the changes were already tracked by PyPads
                branch, diff = check_existing_branches(repo, ref=orig_branch)

                if not branch:
                    logger.warning("Stashing, branching out, "
                                   "committing, reverting back and unstashing...")
                    # branch out, apply the stashed changes and commit
                    branch = "PyPads/{}".format(run.info.run_id)
                    repo.git.checkout(orig_branch, b=branch)

                    repo.git.stash('apply')
                    add_files(repo)
                    repo.git.commit(message=message)

                    # create the tag with diff for this branch
                    diff_raw = repo.git.diff('master', '--raw')
                    diff = repo.git.diff('master')
                    repo.create_tag(path=branch, message=diff_raw)
                    _hash = repo.head.object.hexsha
                else:
                    logger.warning("Using already existing pypads branch " + branch.name)
                    _hash = branch.object.hexsha
                    branch = branch.name

                # Log the commit hash, branch and diff
                pads.api.set_tag("pypads.git.branch", branch)
                pads.api.set_tag("pypads.source.git.commit", _hash)
                pads.api.set_tag("pypads.source.git.diff", diff)

                # checkout to the master branch
                repo.git.checkout(orig_branch)

                # and pop the stash
                repo.git.stash('pop')

            else:
                pads.api.set_tag("pypads.git.branch", orig_branch)
                pads.api.set_tag("pypads.source.git.commit", repo.head.object.hexsha)

        except Exception as e:
            logger.warning("Checking the index of your repository failed due to %s" % str(e))


def check_existing_branches(repo, ref='HEAD'):
    # get the branches created by PyPads
    branches = {b.name: b for b in repo.branches if "PyPads" in b.name}

    # get the tags saved by PyPads
    tags = [t.tag for t in repo.tags if 'PyPads' in t.path]
    try:
        for tag in tags:
            diff = repo.git.diff(ref, tag.tag, '--raw')
            if diff == tag.message:
                print("found a branch with the same changes")
                return branches.get(tag.tag), repo.git.diff(ref, tag.tag)
    except Exception as e:
        logger.info("Checking existing branches failed due to %s" % str(e))
        return None, None
    return None, None


def add_files(repo):
    untracked_files = repo.untracked_files
    if untracked_files and len(untracked_files) > 0:
        repo.index.add(untracked_files)
    if len([item.a_path for item in repo.index.diff(None)]) > 0:
        repo.git.add(A=True)



from .visitor import CombineVisitor

GeneralVisitor = {"__doc__": "docs"}


def generalize_visitor(visitor):
    """Add default fields to the visitor, e.g. the documentation. See GeneralVisitor"""
    return CombineVisitor([GeneralVisitor, visitor])



import sys

from loguru import logger as log

logger = log


class LoggerManager:
    """
    Class to manage loguru handlers for pypads. Loguru doesn't give an extensive api to query already existing loggers
    and therefore we need to manage a history ourselves. Thus to joblib and other parallel processes need us to clear
    the loggers out of the locals before returning we also allow for temporarily deleting loggers.
    """

    def __init__(self):
        # remove preconfigured handler
        try:
            logger.remove(0)
        except Exception as e:
            logger.warning(e)
        self._add_history = {}
        self._removed = []
        self.add_default_logger()

    def add_default_logger(self):
        self.add(sys.stdout, filter="pypads", level="INFO")
        self.add(sys.stderr, filter="pypads", level="INFO")

    def add(self, *args, **kwargs):
        lid = logger.add(*args, **kwargs)
        self._add_history[lid] = (args, kwargs)
        return lid

    def temporary_remove(self):
        for k in list(self._add_history):
            try:
                self.remove(k)
                self._removed.append(k)
            except Exception as e:
                pass

    def add_loggers_from_history(self):
        for k in self._add_history.values():
            try:
                logger.remove(k)
            except Exception as e:
                pass
            self.add(*k[0], **k[1])
            self._removed.remove(k)

    def remove(self, lid=None):
        if lid:
            logger.remove(lid)
            del self._add_history[lid]
        else:
            logger.remove()
            self._add_history = {}


logger_manager = LoggerManager()



import time
from collections import OrderedDict

from pypads import logger
from pypads.logging_util import WriteFormats


class TimingDefined(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def timed(f):
    start = time.time()
    ret = f()
    elapsed = time.time() - start
    return ret, elapsed


def get_logger_times():
    from pypads.pypads import get_current_pads
    pads = get_current_pads()

    timings: OrderedDict = pads.cache.run_get("timings")
    aggregated_times = {}
    for k, v in timings.items():
        splits = k.split(".")
        # Only loggers and not calls themselves
        if splits[-1] == "__pre__" or splits[-1] == "__post__":
            log_function = splits[-2]
            if log_function not in aggregated_times:
                aggregated_times[log_function] = 0
            aggregated_times[log_function] += float(v[1])

    out = ""
    for k, v in sorted(aggregated_times.items(), key=lambda x: -x[1]):
        out += k + " elapsed " + str(v) + "s.\n"
    return out


def print_timings(pads, *args, **kwargs):
    timings: OrderedDict = pads.cache.run_get("timings")
    out = ""
    for k, v in timings.items():
        tabs = ""
        # depth
        for i in range(1, v[0]):
            tabs += "\t"
        out += tabs + str(k) + ": " + str(v[1]) + "\n"
    pads.api.log_mem_artifact("timings", out, write_format=WriteFormats.text.text)
    pads.api.log_mem_artifact("loggers", get_logger_times(), write_format=WriteFormats.text.text)


def add_run_time(log_function, name, time):
    from pypads.pypads import get_current_pads
    pads = get_current_pads()

    pads.api.register_post_fn("timings", print_timings)

    if not pads.cache.run_exists("timings"):
        pads.cache.run_add("timings", OrderedDict())

    timings: OrderedDict = pads.cache.run_get("timings")
    if name not in timings:
        timings[name] = (pads.call_tracker.call_depth(), time)
        logger.debug(name + " done after: " + str(time) + "s")
    else:
        raise TimingDefined("Timing already defined for " + name)

    # # Add to logger time
    # if logger:
    #     if not pads.cache.run_exists("logger_times"):
    #         pads.cache.run_add("logger_times", OrderedDict())
    #     logger_times: OrderedDict = pads.cache.run_get("logger_times")
    #     if logger.__class__.__name__ not in logger_times:
    #         logger_times[logger.__class__.__name__] = [time]
    #     else:
    #         logger_times[logger.__class__.__name__].append(time)



"""
General visitors, that can be combined recursively to extract arbitrary data from any python object.
"""

import abc
import importlib
import inspect
from abc import abstractmethod

from .mappings import type_mappings
from .parameter import Parameter
from .schema import SchemaMismatch


class Visitor(abc.ABC):
    """
    The base class of any Visitor to inspect python objects and extract the setup information.
    For every value there will also be the access string of the item in the original object stored.
    """

    def __init__(self, schema=None):
        """
        Default implementation of constructor.
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__()
        self.schema = schema

    def __call__(self, object, result=None, path=""):
        """
        Make a visitor callable. Calls extract, but also verifies if the result fits the provided schema, if any.
        :param object: see extract
        :param result: (optional) the output-dictionary
        :param path: (optional) the path to keep track of recursive calls
        :return: a tuple containing the result of extract and the schema of the visitor
        """
        if result is None:
            result = {}
        self.extract(object, result, path)
        if self.schema is not None:
            valid = self.schema.verify(result)
            if not valid[0]:
                raise SchemaMismatch(valid[1])
        return result, self.schema

    @abstractmethod
    def extract(self, object, result, path=""):
        """
        (abstract) Must be overridden by subclasses to implement custom extraction behaviour.
        :param object: the object of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-keywords of the template
        """
        return

    @staticmethod
    def applyVisitor(object, result, template, path):
        """
        Applies the Visitor to the object depending on the type of template and stores the result in the result variable:
        - dict: the DictVisitor will be applied to the object using that dict as template
        - tuple: the TupleVisitor will be applied to the object using that tuple as template
        - str: the value of the object will be stored in the result dict using the value of the string as path in the result dict
        - callable: the template will get called with the object and a reference to the result-dictionary as arguments
        - None: no information will be extracted
        """
        if type(template) is str:
            elements = template.split(".")
            last_element = elements[-1]
            elements = elements[:-1]
            target_dict = result
            for e in elements:
                if not e in target_dict:
                    target_dict[e] = {}
                target_dict = target_dict[e]

            if last_element in target_dict:
                print("Warning: Parameter '" + template + "' extracted multiple times! Last value will be used.")
            target_dict[last_element] = Parameter(object, {"path": path})
        elif type(template) is dict:
            DictVisitor(template).extract(object, result, path)
        elif type(template) is tuple:
            TupleVisitor(template).extract(object, result, path)
        elif callable(template):
            template(object, result, path)
        elif template is None:
            pass
        else:
            raise TypeError("Template contains value of unexpected Type: " + str(type(template)))

        return result


class DictVisitor(Visitor):
    """
    A Visitor-implementation to inspect library-specific experiment setups and extract the setup information by using a template-dictionary.
    For each key in the dicitionary the corresponding value will be applied as visitor to a member of object by the same name as key.
    """

    def __init__(self, template, schema=None):
        """
        :param template: template dictionary describing the experiment hierarchy
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__(schema)
        self.template = template

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Returns the result of a call to extract_rec with the given arguments and the stored template.
        :param object: the object of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-keywords of the template
        """
        return DictVisitor.extract_rec(object, result, self.template, path)

    @staticmethod
    def extract_rec(object, result, template, path):
        """
        Stores all information of object extracted using template in the dict result.
        :param object: an object that is part of the experiment description
        :param result: result dictionary used for recursion
        :param template: template used for recursion
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-keywords of the template
        """
        # Check if object is a dict or a class object
        if type(object) is dict:
            getter = lambda k: object[k]
        elif hasattr(object, "__dict__"):
            getter = lambda k: getattr(object, k)
        else:
            raise TypeError("Inspected object has unsupported Type: " + str(type(object)))

        for k in template:
            Visitor.applyVisitor(getter(k), result, template[k], path + "." + k)
            # print("Warning: Parameter '" + k + "' could not be found in object " + repr(object) + ".")

        return result


class ListVisitor(Visitor):
    """
    A Visitor-implementation to inspect library-specific experiment setups and extract the setup information of a list by using a template on every object of the list.
    The template is represented by a dict, where a value can be of any of the types supported by ExperimentVisitor.applyVisitor.
    """

    def __init__(self, prefix, visitor, schema=None):
        """
        :param prefix: name, that is used to store the list of extracted values
        :param visitor: template visitor describing the hierarchy for the elements of the list
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__(schema)
        self.prefix = prefix
        self.visitor = visitor

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Inserts the list of dictionaries,
         each obtained by calling applyVisitor on the respective element in the input list, into result with the key given by self.prefix.
        :param object: the input-list of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: the result dictionary
        """
        if self.prefix in result:
            print("Warning: List '" + path + "." + self.prefix + "' extended by multiple runs.")
        else:
            result[self.prefix] = []
        result[self.prefix].extend(
            [self.applyVisitor(object[i], {}, self.visitor, path + "[" + str(i) + "]") for i in range(len(object))])
        return result


class TupleVisitor(Visitor):
    """
    A Visitor-implementation to inspect library-specific experiment setups and extract the setup information of a tuple-like by applying a template-tuple pairwise on the objects of the tuple.
    The template is represented by a tuple, where a value can be of any of the types supported by ExperimentVisitor.applyVisitor.
    """

    def __init__(self, visitors, schema=None):
        """
        :param visitors: template tuple describing the visitors for the elements in the input-tuple
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__(schema)
        self.visitors = visitors

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Applies pairwise the template-visitor to the elements of the input-object.
        :param object: the input-tuple of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-information
        """
        if len(object) != len(self.visitors):
            print("Warning: Tuple size doesn't match!")
        for i in range(min(len(object), len(self.visitors))):
            self.applyVisitor(object[i], result, self.visitors[i], path + "[" + str(i) + "]")
        return result


class SelectVisitor(Visitor):
    """
    A Visitor-implementation to inspect library-specific experiment setups and select the correct visitor using a decision-dict.
    The Dictionary contains pairs of classes and visitors. A visitor is selected if the object is an instance of the class.
    """

    def __init__(self, visitors, schema=None):
        """
        :param visitors: a dict, that maps base-classes to visitor-objects
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__(schema)
        self.visitors = visitors

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Selects a visitor according to the base class of object. If no class matches the visitor to the key None is used, if available.
        :param object: the input-tuple of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-information
        """
        visitor = None
        for k, v in self.visitors.items():
            if k:
                if isinstance(k, str):
                    splits = k.rsplit('.', 1)
                    module = importlib.import_module(splits[0])
                    k = getattr(module, splits[-1])
                if inspect.isclass(k) and isinstance(object, k):
                    visitor = v
                    break
        if visitor is None:
            if None in self.visitors:
                visitor = self.visitors[None]
        if visitor is None:
            raise TypeError("Unsupported object encountered: " + str(type(object)))
        return self.applyVisitor(object, result, visitor, path)


class CombineVisitor(Visitor):
    """
    A Visitor-implementation to combine multiple visitor-objects.
    """

    def __init__(self, visitors):
        """
        :param visitors: a list, containing all visitors
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__()
        self.visitors = visitors

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Applies all visitors to object.
        :param object: the input-tuple of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-information
        """
        for visitor in self.visitors:
            self.applyVisitor(object, result, visitor, path)
        return result


class ConstantVisitor(Visitor):
    """
    A Visitor-implementation that inserts a constant dict of values into the result-dict.
    """

    def __init__(self, values, schema=None):
        """
        :param values: the constant values
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__(schema)
        self.values = values

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Adds the items in self.values to result.
        :param object: the input-tuple of which information are to be extracted
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-information
        """
        for k in self.values:
            self.applyVisitor(self.values[k], result, k, path)
        return result


class SubpathVisitor(Visitor):
    """
    A Visitor that creates and steps into a subpath of the result dict without changing object and applies its visitor on the object in the new path.
    """

    def __init__(self, subpath, visitor, schema=None):
        """
        :param subpath: the subpath to create
        :param visitor: the visitor
        :param schema: (optional) the expected schema of the extracted paramters
        """
        super().__init__(schema)
        self.subpath = subpath.split(".")
        self.visitor = visitor

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-interface. Creates its subpath in result and calls applyVisitor with its given visitor on object.
        If any element of the path ends with '[]', a list, which will be created if not existent in the subobject with the given key, will be appended a new dicitionary.
        otherwise the dicitonary will be inserted directly into the subobject, if not existing.
        :param object: the input-object to be passed directly to applyVisitor
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-information
        """
        subresult = result
        for s in self.subpath:
            if s.endswith("[]"):
                s = s[:-2]
                if s not in subresult:
                    subresult[s] = []
                subresult[s].append({})
                subresult = subresult[s][-1]
            else:
                if s not in subresult:
                    subresult[s] = {}
                subresult = subresult[s]

        return self.applyVisitor(object, subresult, self.visitor, path)


class AlgorithmVisitor(Visitor):
    """
    A Visitor that uses the information given in mapping.json to extract the information of an object.
    """

    def __init__(self, schema=None, param_additionalInformation=[]):
        """
        Creates a new AlgorithmVisitor.
        :param schema: (optional) the expected schema of the extracted paramters
        :param param_additionalInformation: (optional) set of keys to parameters, for which the respective values from mappings.py should be added to the results
        """
        super().__init__(schema)
        self.param_info = param_additionalInformation

    def setParamAdditionalInfo(self, param_additionalInformation):
        """
        Modifies the set of keys to parameters, for which the respective values from mappings.py should be added to the results.
        :param param_additionalInformation: the new set
        """
        self.param_info = param_additionalInformation

    def extract(self, object, result, path=""):
        """
        Implementation of the ExperimentVisitor-Interface. It looks for a mapping of the object-type to an algorithm.
        If there is none a ValueError is thrown. Otherwise all parameters of the algorithm will be extracted.
        :param object: the input-object to be passed directly to applyVisitor
        :param result: a dictionary containing all extracted information
        :param path: used to keep the path-information of the original item
        :return: a dictionary containing all extracted padre-information
        """

        fullName = "Unknown"
        for key, value in inspect.getmembers(object):
            if key.startswith("_pypads_mapping"):
                for mapping in value:
                    fullName = mapping.reference
                    break
                break
        fullName_ = object.__class__.__module__ + "." + object.__class__.__name__ if hasattr(object,
                                                                                             "__module__") else fullName
        if fullName_ in type_mappings:
            description, lib = type_mappings[fullName_]
        elif fullName != fullName_ and fullName in type_mappings:
            description, lib = type_mappings[fullName]
        else:
            raise ValueError("The algorithm described by class '" + fullName +"/" + fullName_ + "' is not registered!")

        result['algorithm'] = Parameter(description['name'], {"path": path})

        params = description.get('hyper_parameters', {})
        result['hyper_parameters'] = {}
        for param_type in params:
            param_list = {}
            result['hyper_parameters'][param_type] = param_list
            for param in params[param_type]:
                value = object.__dict__
                for k in param[lib]['path'].split('.'):
                    value = value[k]
                # resolve
                param_list[param['name']] = Parameter(value, {"path": path + "." + param[lib]['path'],
                                                              **{k: param[k] for k in self.param_info}})

        return result



# original_init_ = Process.__init__
#
#
# def punched_init_(self, group=None, target=None, name=None, args=(), kwargs={}):
#     if target:
#         run = mlflow.active_run()
#         if run:
#             @wraps(target)
#             def new_target(*args, _pypads=None, _pypads_active_run_id=None, _pypads_tracking_uri=None, _pypads_affected_modules=None, **kwargs):
#                 import mlflow
#                 import pypads.base
#                 pypads.base.current_pads = _pypads
#                 mlflow.set_tracking_uri(_pypads_tracking_uri)
#                 mlflow.start_run(run_id=_pypads_active_run_id)
#                 _pypads.activate_tracking(reload_warnings=False, affected_modules=_pypads_affected_modules, clear_imports=True)
#                 out = target(*args, **kwargs)
#                 # TODO find other way to not close run after process finishes
#                 if len(mlflow.tracking.fluent._active_run_stack) > 0:
#                     mlflow.tracking.fluent._active_run_stack.pop()
#                 return out
#
#             target = new_target
#             kwargs["_pypads"] = current_pads
#             kwargs["_pypads_active_run_id"] = run.info.run_id
#             kwargs["_pypads_tracking_uri"] = mlflow.get_tracking_uri()
#             kwargs["_pypads_affected_modules"] = punched_module_names
#     return original_init_(*self, group=group, target=target, name=name, args=args, kwargs=kwargs)
#
#
# Process.__init__ = punched_init_



import os
import pickle
from enum import Enum
from os.path import expanduser

import mlflow
from mlflow.tracking import MlflowClient
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger


def get_base_folder(run=None):
    """
    Get the base folder to log tmp files to. For now it can't be changed. Todo make configurable
    :return:
    """
    run = run if run else mlflow.active_run()
    if run is None:
        raise ValueError("No active run is defined.")
    return os.path.join(expanduser("~"), ".pypads", run.info.experiment_id, run.info.run_id) + os.path.sep


def get_run_folder():
    """
    Get the folder holding the run information.
    :return:
    """
    run = mlflow.active_run()
    if run is None:
        raise ValueError("No active run is defined.")
    # TODO use artifact download if needed or load artifact.
    return os.path.join(mlflow.get_tracking_uri(), run.info.experiment_id, run.info.run_id)


class WriteFormats(Enum):
    pickle = 1
    text = 2


# extract all tags of runs by experiment id
def all_tags(experiment_id):
    client = MlflowClient(mlflow.get_tracking_uri())
    ds_infos = client.list_run_infos(experiment_id)
    for i in ds_infos:
        yield mlflow.get_run(i.run_id).data.tags


def try_read_artifact(file_name):
    # TODO make defensive
    base_path = get_run_folder()
    path = os.path.join(base_path, "artifacts", file_name)
    with open(path, "r") as meta:
        data = meta.readlines()
    return data


def try_write_artifact(file_name, obj, write_format, preserve_folder=True):
    """
    Function to write an artifact to disk.
    :param write_format:
    :param file_name:
    :param obj:
    :return:
    """
    base_path = get_base_folder()
    path = base_path + file_name

    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

    # Functions for the options to write to
    def write_text(p, o):
        with open(p + ".txt", "w+") as fd:
            fd.write(str(o))
            return fd.name

    def write_pickle(p, o):
        try:
            with open(p + ".pickle", "wb+") as fd:
                pickle.dump(o, fd)
                return fd.name
        except Exception as e:
            logger.warning("Couldn't pickle output. Trying to save toString instead. " + str(e))
            return write_text(p, o)

    # Options to write to
    options = {
        WriteFormats.pickle: write_pickle,
        WriteFormats.text: write_text
    }

    # Write to disk
    if isinstance(write_format, str):
        if WriteFormats[write_format]:
            write_format = WriteFormats[write_format]
        else:
            logger.warning("Configured write format " + write_format + " not supported! ")
            return

    path = options[write_format](path, obj)
    if preserve_folder:
        in_folder = os.path.join(base_path, file_name.split(os.sep)[0])
        # Log artifact to mlflow
        if os.path.isdir(in_folder):
            try_mlflow_log(mlflow.log_artifact, in_folder)
        else:
            try_mlflow_log(mlflow.log_artifact, path)
    else:
        try_mlflow_log(mlflow.log_artifact, path)



import os
from _py_abc import ABCMeta
from abc import abstractmethod

from pypads import logger
from pypads.functions.loggers.base_logger import FunctionHolder
from pypads.functions.loggers.mixins import DependencyMixin, OrderMixin, TimedCallableMixin, IntermediateCallableMixin, \
    ConfigurableCallableMixin, DefensiveCallableMixin


class PreRunFunction(DefensiveCallableMixin, IntermediateCallableMixin, FunctionHolder, TimedCallableMixin,
                     DependencyMixin, OrderMixin,
                     ConfigurableCallableMixin):
    """
    This class should be used to define new pre run functions
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, fn=None, **kwargs):
        super().__init__(*args, fn=fn, **kwargs)
        if self._fn is None:
            self._fn = self._call

    @abstractmethod
    def _call(self, pads, *args, **kwargs):
        """
        Function where to add you custom code to execute before starting the run.

        :param pads: the current instance of PyPads.
        """
        return NotImplementedError()

    def __real_call__(self, *args, **kwargs):
        from pypads.pypads import get_current_pads
        return super().__real_call__(get_current_pads(), *args, **kwargs)

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        logger.warning(str(error))


class RunInfo(PreRunFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def _needed_packages():
        return ["pip"]

    def _call(self, pads, *args, **kwargs):
        logger.info("Tracking execution to run with id " + pads.api.active_run().info.run_id)

        # Execute pip freeze
        try:
            # noinspection PyProtectedMember,PyPackageRequirements
            from pip._internal.operations import freeze
        except ImportError:  # pip < 10.0
            # noinspection PyUnresolvedReferences,PyPackageRequirements
            from pip.operations import freeze
        pads.api.log_mem_artifact("pip_freeze", "\n".join(freeze.freeze()))


class RunLogger(PreRunFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _call(self, pads, *args, **kwargs):
        from pypads.base import PypadsApi
        _api: PypadsApi = pads.api

        from pypads.logging_util import get_base_folder
        folder = get_base_folder()

        # TODO loguru has problems with multiprocessing / make rotation configurable etc
        from pypads.pads_loguru import logger_manager
        lid = logger_manager.add(os.path.join(folder, "run_" + _api.active_run().info.run_id + ".log"),
                                 rotation="50 MB",
                                 enqueue=True)

        import glob

        def remove_logger(pads, *args, **kwargs):
            try:
                from pypads.pads_loguru import logger_manager
                logger_manager.remove(lid)
            except Exception:
                pass
            for file in glob.glob(os.path.join(folder, "run_*.log")):
                pads.api.log_artifact(file)

        _api.register_post_fn("logger_" + str(lid), remove_logger)



import atexit
import os
import signal
import subprocess
from sys import platform

from pypads import logger
from pypads.functions.post_run.post_run import PostRunFunction
from pypads.functions.pre_run.pre_run import PreRunFunction
from pypads.logging_util import get_base_folder


class STrace(PreRunFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _call(self, pads, *args, **kwargs):

        file = os.path.join(get_base_folder(), str(os.getpid()) + "_trace.txt")
        proc = None
        if platform == "linux" or platform == "linux2":
            # https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true
            proc = subprocess.Popen(['sudo strace -p ' + str(os.getpid()) + ' &> ' + file], shell=True,
                                    preexec_fn=os.setsid)

        elif platform == "darwin":
            proc = subprocess.Popen(['sudo dtruss -f -p ' + str(os.getpid()) + ' 2> ' + file], shell=True,
                                    preexec_fn=os.setsid)

        elif platform == "win32":
            logger.warning("No tracing supported on windows currently.")

        if proc:
            pads.api.register_post("stop_dtrace_" + str(proc.pid),
                                   DtraceStop(_pypads_proc=proc, _pypads_trace_file=file))
            if proc.poll() == 1:
                logger.warning(
                    "Can't dtruss/strace without sudo rights. To enable tracking allow user to execute dtruss/strace "
                    "without sudo password with polkit or by modifiying visudo - /etc/sudoers:"
                    "username ALL=NOPASSWD: /usr/bin/dtruss. To get the path to dtruss you can use 'which dtruss'. "
                    "Be carefull about allowing permanent sudo rights to dtruss. This might introduce security risks.")

        def safety_hook():
            """
            A None value indicates that the process hasn't terminated yet.
            """
            if proc and proc.poll() is None:
                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
                proc.terminate()

        atexit.register(safety_hook)


class DtraceStop(PostRunFunction):
    def __init__(self, *args, _pypads_proc=None, _pypads_trace_file=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._proc = _pypads_proc
        self._trace_file = _pypads_trace_file

    def _call(self, pads, *args, **kwargs):
        if self._proc and self._proc.poll() is None:
            os.killpg(os.getpgid(self._proc.pid), signal.SIGTERM)
            self._proc.terminate()
        try:
            pads.api.log_artifact(self._trace_file)
        except Exception as e:
            pass



from pypads.functions.pre_run.pre_run import PreRunFunction
from pypads.util import sizeof_fmt, local_uri_to_path


class ISystem(PreRunFunction):

    @staticmethod
    def _needed_packages():
        return ["platform"]

    def _call(self, pads, *args, **kwargs):
        import platform
        uname = platform.uname()
        pads.api.set_tag("pypads.system", uname.system)
        pads.api.set_tag("pypads.system.node", uname.node)
        pads.api.set_tag("pypads.system.release", uname.release)
        pads.api.set_tag("pypads.system.version", uname.version)
        pads.api.set_tag("pypads.system.machine", uname.machine)
        pads.api.set_tag("pypads.system.processor", uname.processor)


class ICpu(PreRunFunction):

    @staticmethod
    def _needed_packages():
        return ["psutil"]

    def _call(self, pads, *args, **kwargs):
        import psutil
        pads.api.set_tag("pypads.system.cpu.physical_cores", psutil.cpu_count(logical=False))
        pads.api.set_tag("pypads.system.cpu.total_cores", psutil.cpu_count(logical=True))
        freq = psutil.cpu_freq()
        pads.api.set_tag("pypads.system.cpu.max_freq", f"{freq.max:2f}Mhz")
        pads.api.set_tag("pypads.system.cpu.min_freq", f"{freq.min:2f}Mhz")


class IRam(PreRunFunction):

    @staticmethod
    def _needed_packages():
        return ["psutil"]

    def _call(self, pads, *args, **kwargs):
        import psutil
        memory = psutil.virtual_memory()
        pads.api.set_tag("pypads.system.memory.total", sizeof_fmt(memory.total))
        swap = psutil.swap_memory()
        pads.api.set_tag("pypads.system.swap.total", sizeof_fmt(swap.total))


class IDisk(PreRunFunction):

    @staticmethod
    def _needed_packages():
        return ["psutil"]

    def _call(self, pads, *args, **kwargs):
        import psutil
        # see https://www.thepythoncode.com/article/get-hardware-system-information-python
        path = local_uri_to_path(pads._uri)
        disk_usage = psutil.disk_usage(path)
        pads.api.set_tag("pypads.system.disk.total", sizeof_fmt(disk_usage.total))


class IPid(PreRunFunction):

    @staticmethod
    def _needed_packages():
        return ["psutil"]

    def _call(self, pads, *args, **kwargs):
        import psutil
        import os
        pid = os.getpid()
        process = psutil.Process(pid=pid)
        pads.api.set_tag("pypads.system.process.id", pid)
        pads.api.set_tag("pypads.system.process.cwd", process.cwd())
        pads.api.set_tag("pypads.system.process.cpu_usage", str(process.cpu_percent()) + "%")
        pads.api.set_tag("pypads.system.process.memory_usage", str(process.memory_percent()) + "%")


class ISocketInfo(PreRunFunction):

    def _call(self, pads, *args, **kwargs):
        import socket
        pads.api.set_tag("pypads.system.hostname", socket.gethostname())
        pads.api.set_tag("pypads.system.ip-address", socket.gethostbyname(socket.gethostname()))


class IMacAddress(PreRunFunction):

    def _call(self, pads, *args, **kwargs):
        import re, uuid
        pads.api.set_tag("pypads.system.macaddress", ':'.join(re.findall('..', '%012x' % uuid.getnode())))

# def inetw(pads):
#     if is_package_available("psutil"):
#         import psutil
#         # get net stats
#
#     else:
#         logger.warning("To track network usage you need to install psutil.")



import traceback
from _py_abc import ABCMeta
from abc import abstractmethod

import mlflow

from pypads import logger
from pypads.functions.analysis.call_tracker import LoggingEnv
from pypads.functions.analysis.time_keeper import TimingDefined, add_run_time
from pypads.functions.loggers.mixins import DependencyMixin, DefensiveCallableMixin, TimedCallableMixin, \
    IntermediateCallableMixin, NoCallAllowedError, OrderMixin, ConfigurableCallableMixin


class PassThroughException(Exception):
    """
    Exception to be passed from _pre / _post and not be caught by the defensive logger.
    """

    def __init__(self, *args):
        super().__init__(*args)


class FunctionHolder(TimedCallableMixin):
    """
    Holds the given function in a timed callable.
    """

    def __init__(self, *args, fn, **kwargs):
        super().__init__(*args, **kwargs)
        self._fn = fn

    @property
    def fn(self):
        return self._fn

    def __real_call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


class LoggingExecutor(DefensiveCallableMixin, FunctionHolder, ConfigurableCallableMixin):
    __metaclass__ = ABCMeta
    """
    Pre or Post executor for the logging function.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        try:
            raise error
        except TimingDefined:
            pass
        except NotImplementedError:

            # Ignore if only pre or post where defined
            return None, 0
        except (NoCallAllowedError, PassThroughException) as e:

            # Pass No Call Allowed Error through
            raise e
        except Exception as e:

            # Catch other exceptions for this single logger
            try:
                mlflow.set_tag("pypads_failure", str(error))
                logger.error(
                    "Tracking failed for " + str(_pypads_env.call) + " with: " + str(error))
            except Exception as e:
                logger.error("Tracking failed for " + str(_pypads_env.call.call_id.instance) + " with: " + str(error))
            return None, 0


# noinspection PyBroadException
class LoggingFunction(DefensiveCallableMixin, IntermediateCallableMixin, DependencyMixin, OrderMixin):
    """
    This class should be used to define new custom loggers. The user has to define __pre__ and/or __post__ methods
    depending on the specific use case.

    :param static_parameters: dict, optional, static parameters (if needed) to be used when logging.

    .. note:: It is not recommended to change the __call_wrapped__ method, only if really needed.

    """
    __metaclass__ = ABCMeta

    def __init__(self, *args, static_parameters=None, **kwargs):
        super().__init__(*args, **kwargs)
        if static_parameters is None:
            static_parameters = {}
        self._static_parameters = static_parameters

        if not hasattr(self, "_pre"):
            self._pre = LoggingExecutor(fn=self.__pre__)
        if not hasattr(self, "_post"):
            self._post = LoggingExecutor(fn=self.__post__)

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        try:
            raise error
        except NoCallAllowedError as e:

            # Call next wrapped callback if no call was allowed due to the settings or environment
            _pypads_hook_params = {**self._static_parameters, **_pypads_env.parameter}
            return self.__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=args, _kwargs=kwargs,
                                         **_pypads_hook_params)
        except Exception as e:
            logger.error("Logging failed for " + str(self) + ": " + str(error) + "\nTrace:\n" + traceback.format_exc())

            # Try to call the original unwrapped function if something broke
            original = _pypads_env.call.call_id.context.original(_pypads_env.callback)
            if callable(original):
                try:
                    logger.error("Trying to recover from: " + str(e))
                    out = original(ctx, *args, **kwargs)
                    logger.success("Succeeded recovering on error : " + str(e))
                    return out
                except TypeError as e:
                    logger.error("Recovering failed due to: " + str(
                        e) + ". Trying to call without passed ctx. This might be due to an error in the wrapping.")
                    out = original(*args, **kwargs)
                    logger.success("Succeeded recovering on error : " + str(e))
                    return out
            else:

                # Original function was not accessiblete
                raise Exception("Couldn't fall back to original function for " + str(
                    _pypads_env.call.call_id.context.original_name(_pypads_env.callback)) + " on " + str(
                    _pypads_env.call.call_id.context) + ". Can't recover from " + str(error))

    def __pre__(self, ctx, *args, _pypads_env, _args, _kwargs, **kwargs):
        """
        The function to be called before executing the log anchor. the value returned will be passed on to the __post__
        function as **_pypads_pre_return**.


        :return: _pypads_pre_return
        """
        raise NotImplementedError()

    def __post__(self, ctx, *args, _pypads_env, _pypads_pre_return, _pypads_result, _args, _kwargs, **kwargs):
        """
        The function to be called after executing the log anchor.

        :param _pypads_pre_return: the value returned by __pre__.
        :param _pypads_result: the value returned by __call_wrapped__.

        :return: the wrapped function return value
        """
        raise NotImplementedError()

    def _extract_runtime(self, out, _pypads_env, label):
        if type(out) is tuple and len(out) is 2:
            _return = out[0]
            time = out[1]
            if time != 0:
                add_run_time(self, str(_pypads_env.call) + "." + self.__class__.__name__ + "." + label, time)

    def __real_call__(self, ctx, *args, _pypads_env: LoggingEnv, **kwargs):
        # Add the static parameters to our passed parameters
        _pypads_hook_params = {**self._static_parameters, **_pypads_env.parameter}

        _pre_result = None
        try:

            _pre_result = self._pre(ctx, _pypads_env=_pypads_env, _args=args, _kwargs=kwargs,
                            **_pypads_hook_params)
            self._extract_runtime(_pre_result, _pypads_env, "__pre__")
        except TimingDefined:
            pass
        _return = self.__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=args, _kwargs=kwargs, **_pypads_hook_params)

        try:
            out = self._post(ctx, _pypads_pre_return=_pre_result, _pypads_result=_return, _pypads_env=_pypads_env,
                             _args=args, _kwargs=kwargs, **_pypads_hook_params)
            self._extract_runtime(out, _pypads_env, "__post__")
        except TimingDefined:
            pass
        return _return

    def __call_wrapped__(self, ctx, *args, _pypads_env: LoggingEnv, _args, _kwargs, **_pypads_hook_params):
        """
        The real call of the wrapped function. Be carefull when you change this.
        Exceptions here will not be catched automatically and might break your workflow. The returned value will be passed on to __post__ function.

        :return: _pypads_result
        """
        _return, time = FunctionHolder(fn=_pypads_env.callback)(*_args, **_kwargs)
        try:
            add_run_time(None, str(_pypads_env.call), time)
        except TimingDefined as e:
            pass
        return _return



import os
import threading
from collections.__init__ import OrderedDict

from pypads import logger
from pypads.autolog.wrapping.base_wrapper import Context


class FunctionReference:

    def __init__(self, _pypads_context: Context, _pypads_wrappee):
        self._context = _pypads_context
        self._function = _pypads_wrappee
        self._real_context = None
        self._function_type = None

        if self.is_wrapped():
            self._function = self.context.container.__dict__[self._function.__name__]

    @property
    def context(self):
        return self._context

    @property
    def wrappee(self):
        return self._function

    def real_context(self):
        """
        Find where the accessor function was defined
        :return:
        """

        # Return if already found
        if self._real_context:
            return self._real_context
        self._real_context = self._context.real_context(self._function.__name__)
        return self._real_context

    def function_type(self):
        """
        Get the function type of the accessor function.
        :return:
        """

        # Return if already found
        if self._function_type:
            return self._function_type

        if self.context.is_module():
            function_type = "staticmethod"
        else:
            # Get the function type (Method, unbound etc.)
            try:
                real_ctx = self.real_context()
                if real_ctx is None:
                    raise ValueError("Couldn't find real context.")
                function_type = type(real_ctx.get_dict()[self._function.__name__])
            except Exception as e:
                logger.warning("Couldn't get function type of '" + str(self._function.__name__) + "' on '" + str(
                    self.real_context()) + ". Omit logging. " + str(e))
                return None

            # TODO Can we find less error prone ways to get the type of the given fn?
            # Delegate decorator of sklearn obfuscates the real type.
            # if is_package_available("sklearn"):
            #     from sklearn.utils.metaestimators import _IffHasAttrDescriptor
            #     if function_type == _IffHasAttrDescriptor:
            if str(function_type) == "<class 'sklearn.utils.metaestimators._IffHasAttrDescriptor'>":
                function_type = "wrapped"
                self._function = self.real_context().get_dict()[self._function.__name__]

            # Set cached result
            self._function_type = function_type
        return function_type

    def is_static_method(self):
        return "staticmethod" in str(self.function_type())

    def is_function(self):
        return "function" in str(self.function_type())

    def function_name(self):
        return self.wrappee.__name__

    def is_class_method(self):
        return "classmethod" in str(self.function_type())

    def is_wrapped(self):
        return "wrapped" in str(self.function_type())

    @property
    def function_id(self):
        if hasattr(self.wrappee, "__name__"):
            if hasattr(self.context, self.wrappee.__name__):
                return id(getattr(self.context, self.wrappee.__name__))
        return str(id(self.context)) + "." + str(id(self.wrappee))


class CallAccessor(FunctionReference):

    def __init__(self, instance, _pypads_context, _pypads_wrappee):
        super().__init__(_pypads_context, _pypads_wrappee)
        self._instance = instance

    @property
    def instance(self):
        return self._instance

    @property
    def instance_id(self):
        return id(self.instance)

    @classmethod
    def from_function_reference(cls, function_reference: FunctionReference, instance):
        return CallAccessor(instance, function_reference.context, function_reference.wrappee)

    def is_call_identity(self, other):
        if other.is_class_method() or other.is_static_method() or other.is_wrapped():
            if other.context == self.context:
                if other.wrappee.__name__ == self.wrappee.__name__:
                    return True
        if other.is_function():
            if other.instance_id == self.instance_id:
                if other.wrappee.__name__ == self.wrappee.__name__:
                    return True


# class CallMapping(CallAccessor):
#
#     def __init__(self, instance, _pypads_context, _pypads_wrappee, mapping):
#         super().__init__(instance, _pypads_context, _pypads_wrappee)
#         self._mapping = mapping
#
#     @property
#     def mapping(self):
#         return self._mapping


class CallId(CallAccessor):

    def __init__(self, instance, _pypads_context,
                 _pypads_wrappee, instance_number, call_number):
        super().__init__(instance, _pypads_context, _pypads_wrappee)
        self._process = os.getpid()
        self._thread = threading.get_ident()
        self._instance_number = instance_number
        self._call_number = call_number

    @classmethod
    def from_accessor(cls, accessor: CallAccessor, instance_number, call_number):
        return CallId(accessor.instance, accessor.context, accessor.wrappee, instance_number, call_number)

    @property
    def process(self):
        return self._process

    @property
    def thread(self):
        return self._thread

    @property
    def instance_number(self):
        return self._instance_number

    @property
    def call_number(self):
        return self._call_number

    def to_parent_folder(self):
        return os.path.join("process_" + str(self._process) + str(self._thread))

    def to_folder(self):
        return os.path.join(*self.to_fragements())

    def __str__(self):
        return ".".join(self.to_fragements())

    def to_fragements(self):
        return ("process_" + str(self._process), "thread_" + str(self._thread),
                "context_" + self.context.container.__name__,
                "instance_" + str(
                    self.instance_number), "function_" + self.wrappee.__name__, "call_" + str(self._call_number))


class Call:

    def __init__(self, call_id: CallId):
        self._call_id = call_id
        self._finished = False
        self._active_hooks = set()

    @property
    def call_id(self):
        return self._call_id

    @property
    def finished(self):
        return self._finished

    @property
    def active_hooks(self):
        return self._active_hooks

    def finish(self):
        self._finished = True

    def add_hook(self, hook):
        self._active_hooks.add(hook)

    def has_hook(self, hook):
        return hook in self._active_hooks

    def remove_hook(self, hook):
        self._active_hooks.remove(hook)

    def __str__(self):
        return self.call_id.__str__()

    def to_folder(self):
        return self.call_id.to_folder()

    # def __getstate__(self):
    #     """
    #     Overwrite standard pickling by excluding the functions
    #     :return:
    #     """
    #     # can't pickle call_ids here
    #     self.call_id_fragments = str(self.call_id.to_fragements())
    #     state = self.__dict__.copy()
    #     del state["_call_id"]
    #     return state
    #
    # def __setstate__(self, state):
    #     self.__dict__.update(state)
    #     state["_call_id"] = None
    #     # can we rebuild call_id?
    #     return state


class LoggingEnv:

    def __init__(self, mapping, hook, parameter, callback, call: Call):
        self._call = call
        self._callback = callback
        self._hook = hook
        self._mapping = mapping
        self._parameter = parameter

    @property
    def call(self):
        return self._call

    @property
    def callback(self):
        return self._callback

    @property
    def hook(self):
        return self._hook

    @property
    def mapping(self):
        return self._mapping

    @property
    def parameter(self):
        return self._parameter


class CallTracker:
    """
    This class tracks the number of execution per instance of an object.
    """

    def __init__(self, pads):
        self._pads = pads
        self._call_stack = []

    def instance_call_number(self, accessor):
        function_calls: OrderedDict = self.function_call_dict(accessor.function_id)
        instance_id = accessor.instance_id

        call_items = list(function_calls.items())
        for instance_number in range(0, len(call_items)):
            # noinspection PyUnresolvedReferences
            key, value = call_items[instance_number]
            if key == instance_id:
                return instance_number
        return 0

    @property
    def call_stack(self):
        return self._call_stack

    def call_depth(self):
        return len(self._call_stack)

    def call_number(self, accessor: CallAccessor):
        """
        Get the number of calls of given call accessor
        :param accessor:
        :return:
        """
        return len(self.calls(accessor))

    def make_call_id(self, accessor: CallAccessor) -> CallId:
        """
        Returns the current call id. The id is built from: process_id, thread_id, defining_ctx_name, self_number/id,
        wrapped_fn_name, call_number
        :param accessor:
        :return:
        """
        return CallId(accessor.instance, accessor.context, accessor.wrappee, self.instance_call_number(accessor),
                      self.call_number(accessor))

    def current_call_number(self):
        """
        Get the current call number
        :return:
        """
        call = self._call_stack[-1]
        return self.call_number(call.accessor)

    def current_call(self):
        """
        Get the call_id of the current call
        :return:
        """
        return self._call_stack[-1] if len(self._call_stack) > 0 else None

    def current_process(self):
        return str(self._call_stack[-1].call_id.process) + "." + str(self._call_stack[-1].call_id.thread)

    def call_objects(self):
        """
        Get all call objects which are stored in our tracker.
        :return:
        """
        # Add call_objects if not exists in cache
        if not self._pads.cache.run_exists("call_objects"):
            self._pads.cache.run_add("call_objects", {})
        return self._pads.cache.run_get("call_objects")

    def function_call_dict(self, function_id) -> OrderedDict:
        call_objects = self.call_objects()
        if function_id not in call_objects:
            call_objects[function_id] = OrderedDict()
        return call_objects[function_id]

    def calls(self, accessor: CallAccessor):
        """
        Get all calls of given call accessor.
        :param accessor:
        :return:
        """
        instance_id = accessor.instance_id
        function_id = accessor.function_id

        function_calls = self.function_call_dict(function_id)
        if instance_id not in function_calls:
            function_calls[instance_id] = []

        return function_calls[instance_id]

    def has_call_identity(self, accessor: CallAccessor):
        for stored in self._call_stack:
            if stored.is_call_identity(accessor):
                return True
        return False

    def add(self, call: Call):
        """
        When a new call is triggered execute that function. This should be the first logging function called in each
        logging function stack.
        :return: A dict for holding information about the call.
        """
        self._call_stack.append(call)

        calls = self.calls(call.call_id)
        calls.append(call)
        return call

    def finish(self, call):
        if call in self._call_stack:
            call.finish()
            self._call_stack.remove(call)
            # TODO clear memory in call_objects?
        else:
            logger.error("Tried to finish call which is not on the stack. " + str(call))


def add_call(accessor):
    from pypads.pypads import get_current_pads
    pads = get_current_pads()
    call = Call(pads.call_tracker.make_call_id(accessor))
    return pads.call_tracker.add(call)


def finish_call(call):
    from pypads.pypads import get_current_pads
    pads = get_current_pads()
    return pads.call_tracker.finish(call)



from pypads.functions.analysis.validation.visitors.experimentschema import default_pipeline_schema
from pypads.functions.analysis.validation.visitors.generalvisitor import generalize_visitor
from pypads.functions.analysis.validation.visitors.visitor import ListVisitor, AlgorithmVisitor, SubpathVisitor, \
    SelectVisitor

SciKitPipelineVisitor = {
                "steps":
                    ListVisitor("steps",
                                (
                                              None,
                                              generalize_visitor(AlgorithmVisitor())
                                          )
                                ),
                "__doc__":
                    "docs"
            }

default_selector = {
    "sklearn.pipeline.Pipeline": SciKitPipelineVisitor,
    None: SubpathVisitor("steps[]", generalize_visitor(AlgorithmVisitor()))
}

default_visitor = SelectVisitor(default_selector, default_pipeline_schema)



import mlflow
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger
from pypads.functions.analysis.call_tracker import LoggingEnv
from pypads.functions.analysis.validation.generic_visitor import default_visitor
from pypads.functions.loggers.base_logger import LoggingFunction


class Parameters(LoggingFunction):

    def __pre__(self, ctx, *args, _pypads_env, **kwargs):
        pass

    def __post__(self, ctx, *args, _pypads_env: LoggingEnv, **kwargs):
        """
        Function logging the parameters of the current pipeline object function call.
        :param ctx:
        :param args:
        :param kwargs:
        :return:
        """
        try:
            # prevent wrapped_class from becoming unwrapped
            visitor = default_visitor(ctx)

            if "model_parameters" in visitor[0]["steps"][0]["hyper_parameters"]:
                for k, v in visitor[0]["steps"][0]["hyper_parameters"]["model_parameters"].items():
                    try:
                        try_mlflow_log(mlflow.log_param, _pypads_env.mapping.reference + "." + k + ".txt", v)
                    except Exception as e:
                        logger.warning("Couldn't track parameter. " + str(e) + " Trying to track with another name.")
                        try_mlflow_log(mlflow.log_param,
                                       str(_pypads_env.call) + "." + k + ".txt", v)

        except Exception as e:
            logger.warning("Couldn't use visitor for parameter extraction on " + str(ctx.__class__) + "Error: " + str(
                e) + ". Omit logging for now.")



from pypads import logger

from pypads.functions.loggers.base_logger import LoggingFunction


class Log(LoggingFunction):
    """
    Function just logging the execution into debug.
    """

    def __pre__(self, ctx, *args, _pypads_env, **kwargs):
        logger.debug("Entered " + str(_pypads_env.call))

    def __post__(self, ctx, *args, _pypads_env, **kwargs):
        logger.debug("Exited " + str(_pypads_env.call))


class LogInit(LoggingFunction):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    def __pre__(self, ctx, *args, **kwargs):
        logger.info("Pypads tracked class " + str(ctx.__class__) + " initialized.")



from pypads import logger

from .mappings import name_mappings
from .parameter import Parameter


class SchemaMismatch(Exception):
    """Exception for schema-mismatch

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class Attribute(object):
    """
    Contains the following information about one attribute in a schema:
        the name,
        the type,
        a flag indicating if the attribute is optional and
        a description
    of the attribute
    """
    def __init__(self, name, desc, optional, type):
        """
        Sets up the information.
        :param name: the name of the attribute
        :param desc: the description of the attribute
        :param optional: indicating if the attribute is optional
        :param type: the type of the attribute
        """
        self.name = name
        self.type = type
        self.optional = optional
        self.desc = desc


class ListAttribute(Attribute):
    """
    An attribute with the type list, that contains a recursive schema to test the lists elements
    """
    def __init__(self, name, desc, optional, schema):
        """
        Sets up the information
        :param name: the name of the attribute
        :param desc: the description of the attribute
        :param optional: indicating if the attribute is optional
        :param schema: the recursive schema
        """
        super().__init__(name, desc, optional, list)
        self.schema = schema

    def verify(self, data, path, keyset):
        """
        Verifies all its elements using self.schema.
        :param exp_schema: the toplevel schema object
        :param data: the list to verify
        :param path: used for recursion. Used for readable error messages.
        :param keyset: used for recursion. Contains the unchecked keys of the data.
        :return: (bool, str) (True, ""), if the data fits. (False, <reason>), otherwise.
        """
        for i in range(len(data)):
            tmp = Schema.verify_static(data[i], self.schema, path + "[" + str(i) + "]")
            if not tmp[0]:
                return tmp
        return (True, "")


class DictAttribute(Attribute):
    """
    An attribute with the type dict, that contains a recursive schema to test the dicts elements
    """

    def __init__(self, schema, name="", desc="", optional=False):
        """
        Sets up the information
        :param schema: the recursive schema
        :param name: the name of the attribute, default=""
        :param desc: the description of the attribute, default=""
        :param optional: indicating if the attribute is optional, default=False
        """
        super().__init__(name, desc, optional, list)
        self.schema = schema

    def verify(self, data, path, keyset):
        """
        Verifies all its elements using self.schema.
        :param data: the list to verify
        :param path: used for recursion. Used for readable error messages.
        :param keyset: used for recursion. Contains the unchecked keys of the data.
        :return: (bool, str) (True, ""), if the data fits. (False, <reason>), otherwise.
        """

        for k in self.schema:
            p = path + "." + k
            if k in data:

                tmp = Schema.verify_static(data[k], self.schema[k], p)
                if not tmp[0]:
                    return tmp
                keyset.remove(k)

            elif not self.schema[k].optional:
                return (False, "Missing Parameter '" + p + "'.")

        return (True, "")


class SelectSchema(object):
    """
    A schema object, that has a decision function to get the correct schema for the dict and verifies it.
    """

    def __init__(self, decision):
        """
        Sets up the decision function
        :param decision: a decision function, that takes a dict and returns a schema dictionary
        """
        self.decision = decision

    def verify(self, data, path, keyset):
        """
        Verifies all its elements using self.schema.
        :param data: the list to verify
        :param path: used for recursion. Used for readable error messages.
        :param keyset: used for recursion. Contains the unchecked keys of the data.
        :return: (bool, str) (True, ""), if the data fits. (False, <reason>), otherwise.
        """
        return Schema.verify_static(data, self.decision(data), path, keyset)


class CombineSchemata(object):
    """
    A schema object, that contains a list of schema-objects, which are all verified against the data.
    """

    def __init__(self, schemata):
        """
        Sets up the decision function
        :param decision: a decision function, that takes a dict and returns a schema dictionary
        """
        self.schemata = schemata

    def verify(self, data, path, keyset):
        """
        Verifies the data using self.schemata.
        :param data: the data to verify
        :param path: used for recursion. Used for readable error messages.
        :param keyset: used for recursion. Contains the unchecked keys of the data.
        :return: (bool, str) (True, ""), if the data fits. (False, <reason>), otherwise.
        """
        for schema in self.schemata:
            tmp = Schema.verify_static(data, schema, path, keyset)
            if not tmp[0]:
                return tmp
        return (True, "")


class Schema(object):
    """
    Describing a schema for experiment parameters.
    """

    def __init__(self, schema):
        """
        constructs the schema with a schema object.
        :param schema: the object describing the schema.
        """
        self.schema = schema

    def verify(self, data):
        """
        Calls ExperimentVisitor.verify_static(data, self.schema).
        :param data: the data to be verified.
        :return: the result of the call to verify_static.
        """
        return Schema.verify_static(data, self.schema)

    @staticmethod
    def verify_static(data, schema, path="", keyset=None):
        """
        Verifies that the given data fits into the defined schema.
        :param data: the data to be checked
        :param schema: the schema to be checked against.
        :param keyset: used for warning-prompts. Contains all unchecked keys. If None it will be created from data
        :param path: (optional) used for recursion. Used for readable error messages.
        :return: (bool, str) (True, ""), if the data fits. (False, <reason-str>), otherwise.
        """

        if keyset is None:
            keyset_was_none = True
            if type(data) is dict:
                keyset = list(data.keys())
            else:
                keyset = []
        else:
            keyset_was_none = False

        if type(data) is list:
            if type(schema) is ListAttribute:
                tmp = schema.verify(data, path, keyset)
            else:
                return (False, "Parameter '" + path + "' has to be of type " + str(
                    schema.type) + " but is of type " + str(list) + ".")

            if not tmp[0]:
                return tmp
        elif type(data) is dict:
            if type(schema) in [SelectSchema, AlgorithmSchema]:
                tmp = schema.verify(data, path, keyset)
            elif type(schema) is dict:
                tmp = DictAttribute(schema).verify(data, path, keyset)
            elif type(schema) is list:
                tmp = CombineSchemata(schema).verify(data, path, keyset)
            elif type(schema) is DictAttribute:
                tmp = schema.verify(data, path, keyset)
            else:
                return (False, "Parameter '" + path + "' has to be of type " + str(
                    schema.type) + " but is of type " + str(dict) + ".")
            if not tmp[0]:
                return tmp
        elif type(data) is Parameter:
            if data.type() is not schema.type:
                return (False,
                        "Parameter '" + path + "' has to be of type " + str(schema.type) + " but is of type " + str(
                            data.type()) + ".")
        else:
            return (False, "Parameter '" + path + "' has unsupported type " + str(type(data)) + ".")

        if keyset_was_none:
            for l in keyset:
                logger.warning("Warning: Attribute '" + path + "." + l + "' is not part of the schema.")

        return (True, "")


class AlgorithmSchema(object):
    """
    Describing the schema for an algorithm given the description in 'mappings.json'.
    """

    def verify(self, data, path, keyset):
        """
        Verifies that data fits the
        :param data: the data to be verified.
        :return: (bool, str) (True, ""), if the data fits. (False, <reason-str>), otherwise.
        """

        if not 'hyper_parameters' in data:
            return (False, "Missing 'hyper_parameters'-field!")

        alg = name_mappings[data['algorithm']]

        for param_type in alg.get('hyper_parameters', []):
            if not param_type in data['hyper_parameters']:
                return (False, "Missing parameter-set '" + param_type + "'!")
            for param in alg['hyper_parameters'][param_type]:
                if not param['name'] in data['hyper_parameters'][param_type]:
                    return (False, "Missing parameter '" + param['name'] + "'!")


        keyset.remove('hyper_parameters')

        return (True, "")



import types
from contextlib import contextmanager
from functools import wraps

from pypads import logger
from pypads.autolog.wrapping.base_wrapper import BaseWrapper, Context
from pypads.functions.analysis.call_tracker import CallAccessor, FunctionReference, add_call, LoggingEnv, finish_call, \
    Call


class FunctionWrapper(BaseWrapper):

    def wrap(self, fn, context: Context, mapping):
        # Only wrap functions not starting with "__"
        if (fn.__name__.startswith("__") or fn.__name__.startswith("_pypads")) and fn.__name__ is not "__init__":
            return fn

        if not context.has_wrap_meta(mapping, fn):
            context.store_wrap_meta(mapping, fn)

            if not context.has_original(fn) or not context.defined_stored_original(fn):
                context.store_original(fn)

            if context.is_class():
                return self._wrap_on_class(fn, context, mapping)
            elif hasattr(context.container, fn.__name__):
                return self._wrap_on_object(fn, context, mapping)
            else:
                logger.warning(str(
                    context) + " is no class and doesn't provide attribute with fn_name. Couldn't access " + str(
                    fn) + " on it.")

    def _wrap_on_object(self, fn, context: Context, mapping):
        # Add module of class to the changed modules
        if hasattr(context.container, "__module__"):
            self._pypads.punched_module_names.add(context.container.__module__)

        # If we are already punched get original function instead of punched function
        if context.has_original(fn):
            fn = context.original(fn)
        else:
            fn = getattr(context.container, fn.__name__)

        # Get and add hooks
        hooks = self._get_hooked_fns(fn, mapping)
        if len(hooks) > 0:
            fn_reference = FunctionReference(context, fn)
            return self.wrap_method_helper(fn_reference=fn_reference, hooks=hooks, mapping=mapping)
        else:
            return fn

    def _wrap_on_class(self, fn, context: Context, mapping):
        fn_name = fn.__name__

        # Find the real defining class
        defining_class = context.real_context(fn.__name__)

        # If there is no defining class we can't wrap on class
        if not defining_class:
            logger.warning("Defining class is None on context '" + str(context) + "' for fn '" + str(
                fn) + "'. Class punching might be lost in newly spawned processes.")
        else:
            # Add the module to the list of modules which where changed
            if hasattr(defining_class.container, "__module__"):
                self._pypads.wrap_manager.module_wrapper.punched_module_names.add(defining_class.container.__module__)

            # Get real fn from defining class
            fn = None
            try:
                fn = getattr(defining_class.container, fn_name)
            except Exception as e:
                context.real_context(fn.__name__)
                logger.warning("Defining class doesn't define our function. Extraction failed: " + str(e))

        # skip wrong extractions
        if not fn or not callable(fn):
            return fn

        # if we have a property we should use fget instead
        if isinstance(fn, property):
            fn = fn.fget

        hooks = self._get_hooked_fns(fn, mapping)
        if len(hooks) > 0:
            fn_reference = FunctionReference(context, fn)
            return self.wrap_method_helper(fn_reference=fn_reference, hooks=hooks, mapping=mapping)

    @contextmanager
    def _make_call(self, instance, fn_reference):
        accessor = CallAccessor.from_function_reference(fn_reference, instance)

        current_call = None
        call = None
        try:
            current_call: Call = self._pypads.call_tracker.current_call()
            if current_call and accessor.is_call_identity(current_call.call_id):
                call = current_call
            else:
                call = add_call(accessor)
            yield call
        finally:
            if call and not current_call == call:
                # print("c:" + str(call))
                # print("cc:" + str(current_call))
                finish_call(call)

    def wrap_method_helper(self, fn_reference: FunctionReference, hooks, mapping):
        """
        Helper to differentiate between functions, classmethods, static methods and wrap them
        :param fn_reference:
        :param hooks:
        :param mapping:
        :return:
        """
        fn = fn_reference.wrappee

        if fn_reference.is_static_method():
            @wraps(fn)
            def entry(*args, _pypads_hooks=hooks, _pypads_mapped_by=mapping, **kwargs):
                logger.debug("Call to tracked static method or function " + str(fn))

                with self._make_call(None, fn_reference) as call:
                    accessor = call.call_id
                    callback = fn

                    # for every hook add
                    if self._is_skip_recursion(accessor):
                        logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                            accessor.wrappee.__name__))
                        out = callback(*args, **kwargs)
                        return out

                    for (h, params) in hooks:
                        c = self._add_hook(h, params, callback, call, mapping)
                        if c:
                            callback = c

                    # start executing the stack
                    out = callback(*args, **kwargs)
                return out
        elif fn_reference.is_function():
            @wraps(fn)
            def entry(_self, *args, _pypads_hooks=hooks, _pypads_mapped_by=mapping, **kwargs):
                # print("Call to tracked class method " + str(fn) + str(id(fn)))
                logger.debug("Call to tracked method " + str(fn))

                with self._make_call(_self, fn_reference) as call:
                    accessor = call.call_id
                    # add the function to the callback stack
                    callback = types.MethodType(fn, _self)

                    # for every hook add
                    if self._is_skip_recursion(accessor):
                        logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                            accessor.wrappee.__name__))
                        out = callback(*args, **kwargs)
                        return out

                    for (h, params) in hooks:
                        c = self._add_hook(h, params, callback, call, mapping)
                        if c:
                            callback = types.MethodType(c, _self)

                    # start executing the stack
                    out = callback(*args, **kwargs)
                return out

        elif fn_reference.is_class_method():
            @wraps(fn)
            def entry(_cls, *args, _pypads_hooks=hooks, _pypads_mapped_by=mapping, **kwargs):
                logger.debug("Call to tracked class method " + str(fn))
                with self._make_call(_cls, fn_reference) as call:
                    accessor = call.call_id
                    # add the function to the callback stack
                    callback = types.MethodType(fn, _cls)

                    # for every hook add
                    if self._is_skip_recursion(accessor):
                        logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                            accessor.wrappee.__name__))
                        out = callback(*args, **kwargs)
                        return out

                    for (h, params) in hooks:
                        c = self._add_hook(h, params, callback, call, mapping)
                        if c:
                            callback = types.MethodType(c, _cls)

                    # start executing the stack
                    out = callback(*args, **kwargs)
                return out

        elif fn_reference.is_wrapped():
            tmp_fn = getattr(fn_reference.context.container, fn.__name__)

            @wraps(tmp_fn)
            def entry(_self, *args, _pypads_hooks=hooks, _pypads_mapped_by=mapping, **kwargs):
                logger.debug("Call to tracked _IffHasAttrDescriptor " + str(fn))
                with self._make_call(_self, fn_reference) as call:
                    accessor = call.call_id
                    # add the function to the callback stack
                    callback = fn.__get__(_self)

                    # for every hook add
                    if self._is_skip_recursion(accessor):
                        logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                            accessor.wrappee.__name__))
                        out = callback(*args, **kwargs)
                        return out

                    for (h, params) in hooks:
                        c = self._add_hook(h, params, callback, call, mapping)
                        if c:
                            callback = types.MethodType(c, _self)

                    # start executing the stack
                    out = callback(*args, **kwargs)
                return out
        else:
            return fn
        fn_reference.context.overwrite(fn.__name__, entry)
        # print("Wrapped " + str(fn) + str(id(fn)))
        return entry

    def _add_hook(self, hook, params, callback, call: Call, mapping):
        # For every hook we defined on the given function in out mapping file execute it before running the code
        if not call.has_hook(hook):
            return self._get_env_setter(_pypads_env=LoggingEnv(mapping, hook, params, callback, call))
        else:
            logger.debug(str(hook) + " is tracked multiple times on " + str(call) + ". Ignoring second hooking.")
            return None

    def _get_env_setter(self, _pypads_env: LoggingEnv):
        """
        This wrapper sets the context
        :return:
        """
        env = _pypads_env
        call: Call = env.call
        cid = call.call_id

        if cid.is_static_method():
            @wraps(cid.wrappee)
            def env_setter(*args, _pypads_env=env, **kwargs):
                logger.debug("Static method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(None, *args, _pypads_env=_pypads_env, **kwargs)

            return env_setter
        elif cid.is_function():
            @wraps(cid.wrappee)
            def env_setter(_self, *args, _pypads_env=env, **kwargs):
                logger.debug("Method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(_self, *args, _pypads_env=_pypads_env, **kwargs)

            return env_setter
        elif cid.is_class_method():
            @wraps(cid.wrappee)
            def env_setter(_cls, *args, _pypads_env=env, **kwargs):
                logger.debug("Class method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(_cls, *args, _pypads_env=_pypads_env,
                                                    **kwargs)

            return env_setter
        elif cid.is_wrapped():
            tmp_fn = getattr(cid.context.container, cid.wrappee.__name__)

            @wraps(tmp_fn)
            def env_setter(_self, *args, _pypads_env=env, **kwargs):
                logger.debug("Method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(_self, *args, _pypads_env=_pypads_env, **kwargs)

            return env_setter
        else:
            raise ValueError("Failed!")

    def _wrapped_inner_function(self, _self, *args, _pypads_env: LoggingEnv, **kwargs):
        """
        Wrapped function logic. This executes all hooks and the function itself.
        :param _self:
        :param args:
        :param kwargs:
        :return:
        """

        env = _pypads_env
        call = env.call
        call.add_hook(env.hook)

        try:
            # check for name collision in parameters
            if set([k for k, v in kwargs.items()]) & set(
                    [k for k, v in env.parameter.items()]):
                logger.warning("Hook parameter is overwriting a parameter in the standard "
                               "model call. This most likely will produce side effects.")

            if env.hook:
                return env.hook(_self, _pypads_env=_pypads_env, *args, **kwargs)
            return env.callback(*args, **kwargs)
        finally:
            call.remove_hook(env.hook)

    def _is_skip_recursion(self, accessor):
        try:
            config = self._pypads.config

            if 'recursion_depth' in config and config['recursion_depth'] is not -1:
                if self._pypads.call_tracker.call_depth() > config['recursion_depth'] + 1:
                    return True
            if 'recursion_identity' in config and config['recursion_identity']:
                if self._pypads.call_tracker.has_call_identity(accessor):
                    return True
            return False
        except Exception as e:
            return False



import os

from pypads.functions.analysis.call_tracker import LoggingEnv
from pypads.functions.loggers.base_logger import LoggingFunction
from pypads.logging_util import try_write_artifact, WriteFormats
from pypads.util import local_uri_to_path, sizeof_fmt


class Cpu(LoggingFunction):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    def _needed_packages(self):
        return ["psutil"]

    def __pre__(self, ctx, *args, _pypads_env: LoggingEnv, **kwargs):
        name = os.path.join(_pypads_env.call.to_folder(), "pre_cpu_usage")
        try_write_artifact(name, _get_cpu_usage(), WriteFormats.text)

    def __call_wrapped__(self, ctx, *args, _pypads_env, _args, _kwargs, **_pypads_hook_params):
        # TODO track while executing instead of before and after
        return super().__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=_args, _kwargs=_kwargs,
                                        **_pypads_hook_params)

    def __post__(self, ctx, *args, _pypads_env: LoggingEnv, **kwargs):
        name = os.path.join(_pypads_env.call.to_folder(), "post_cpu_usage")
        try_write_artifact(name, _get_cpu_usage(), WriteFormats.text)


def _get_cpu_usage():
    import psutil
    cpu_usage = "CPU usage for cores:"
    for i, percentage in enumerate(psutil.cpu_percent(percpu=True)):
        cpu_usage += f"\nCore {i}: {percentage}%"
    cpu_usage += f"\nTotal CPU usage: {psutil.cpu_percent()}%"

    return cpu_usage


class Ram(LoggingFunction):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    def _needed_packages(self):
        return ["psutil"]

    def __pre__(self, ctx, *args, _pypads_env: LoggingEnv, **kwargs):
        name = os.path.join(_pypads_env.call.to_folder(), "pre_memory_usage")
        try_write_artifact(name, _get_memory_usage(), WriteFormats.text)

    def __call_wrapped__(self, ctx, *args, _pypads_env, _args, _kwargs, **_pypads_hook_params):
        # TODO track while executing instead of before and after
        return super().__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=_args, _kwargs=_kwargs,
                                        **_pypads_hook_params)

    def __post__(self, ctx, *args, _pypads_env: LoggingEnv, **kwargs):
        name = os.path.join(_pypads_env.call.to_folder(), "post_memory_usage")
        try_write_artifact(name, _get_memory_usage(), WriteFormats.text)


def _get_memory_usage():
    import psutil
    memory = psutil.virtual_memory()
    swap = psutil.swap_memory()
    memory_usage = "Memory usage:"
    memory_usage += f"\n\tAvailable:{sizeof_fmt(memory.available)}"
    memory_usage += f"\n\tUsed:{sizeof_fmt(memory.used)}"
    memory_usage += f"\n\tPercentage:{memory.percent}%"
    memory_usage += f"\nSwap usage::"
    memory_usage += f"\n\tFree:{sizeof_fmt(swap.free)}"
    memory_usage += f"\n\tUsed:{sizeof_fmt(memory.used)}"
    memory_usage += f"\n\tPercentage:{memory.percent}%"

    return memory_usage


class Disk(LoggingFunction):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    def _needed_packages(self):
        return ["psutil"]

    def __pre__(self, ctx, *args, _pypads_env, **kwargs):
        from pypads.base import PyPads
        from pypads.pypads import get_current_pads
        pads: PyPads = get_current_pads()
        path = local_uri_to_path(pads._uri)
        name = os.path.join(_pypads_env.call.to_folder(), "pre_disk_usage")
        try_write_artifact(name, _get_disk_usage(path), WriteFormats.text)

    def __call_wrapped__(self, ctx, *args, _pypads_env, _args, _kwargs, **_pypads_hook_params):
        # TODO track while executing instead of before and after
        return super().__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=_args, _kwargs=_kwargs,
                                        **_pypads_hook_params)

    def __post__(self, ctx, *args, _pypads_env, **kwargs):
        from pypads.base import PyPads
        from pypads.pypads import get_current_pads
        pads: PyPads = get_current_pads()
        path = local_uri_to_path(pads._uri)
        name = os.path.join(_pypads_env.call.to_folder(), "post_disk_usage")
        try_write_artifact(name, _get_disk_usage(path), WriteFormats.text)


def _get_disk_usage(path):
    import psutil
    # See https://www.thepythoncode.com/article/get-hardware-system-information-python
    disk_usage = psutil.disk_usage(path)
    output_ = "Disk usage:"
    output_ += f"\n\tFree:{sizeof_fmt(disk_usage.free)}"
    output_ += f"\n\tUsed:{sizeof_fmt(disk_usage.used)}"
    output_ += f"\n\tPercentage:{disk_usage.percent}%"
    output_ += f"\nPartitions:"
    partitions = psutil.disk_partitions()
    for partition in partitions:
        output_ += f"\n+Partition1:"
        output_ += f"\n\tDevice:{partition.device}"
        output_ += f"\n\tMountpoint:{partition.mountpoint}"
        output_ += f"\n\tFile system:{partition.fstype}"
        output_ += f"\n\tStats:"
        try:
            partition_usage = psutil.disk_usage(partition.mountpoint)
        except PermissionError:
            output_ += f"\n\t\t Busy!"
            continue
        output_ += f"\n\t\tFree:{partition_usage.free}"
        output_ += f"\n\t\tUsed:{partition_usage.used}"
        output_ += f"\n\t\tPercentage:{partition_usage.percent}%"
    return output_



import inspect
from types import ModuleType

from pypads.autolog.wrapping.base_wrapper import Context
from pypads.autolog.wrapping.class_wrapping import ClassWrapper
from pypads.autolog.wrapping.function_wrapping import FunctionWrapper
from pypads.autolog.wrapping.module_wrapping import ModuleWrapper


def _add_found_class(mapping):
    from pypads.pypads import get_current_pads
    get_current_pads().mapping_registry.add_found_class(mapping)


class WrapManager:

    def __init__(self, pypads):
        self._pypads = pypads
        self._module_wrapper = ModuleWrapper(pypads)
        self._class_wrapper = ClassWrapper(pypads)
        self._function_wrapper = FunctionWrapper(pypads)

    @property
    def module_wrapper(self):
        return self._module_wrapper

    @property
    def class_wrapper(self):
        return self._class_wrapper

    @property
    def function_wrapper(self):
        return self._function_wrapper

    def wrap(self, wrappee, ctx, mapping):
        """
        Wrap given object with pypads functionality
        :param wrappee:
        :param args:
        :param kwargs:
        :return:
        """
        if not str(wrappee).startswith("_pypads") and not str(wrappee).startswith("__"):
            if not isinstance(ctx, Context):
                try:
                    ctx = Context(ctx)
                except ValueError as e:

                    dummy = ModuleType("dummy_module")
                    if inspect.isfunction(wrappee):
                        setattr(dummy, wrappee.__name__, wrappee)
                    ctx = Context(dummy)

            if inspect.ismodule(wrappee):
                return self._module_wrapper.wrap(wrappee, ctx, mapping)

            elif inspect.isclass(wrappee):
                return self._class_wrapper.wrap(wrappee, ctx, mapping)

            elif inspect.isfunction(wrappee):
                return self._function_wrapper.wrap(wrappee, ctx, mapping)
        else:
            return wrappee



class Parameter(object):
    """
    A class representing a single parameter, that was extracted using the ExperimentVisitor.
    It contains the extracted value and the path in the original object to that parameter.
    """

    def __init__(self, value, attributes):
        """
        Constructs a new Parameter.
        :param value: the value of the parameter
        :param attributes: A dictionary containing additional information to that Parameter
        """
        self.value = value
        self.attributes = attributes

    def type(self):
        """
        Returns the type of the value of the parameter.
        :return: the type of the value
        """
        return type(self.value)

    def __ne__(self, other):
        return self.value != other

    def __eq__(self, other):
        return self.value == other

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return "{ value: " + repr(self.value) + ", " + ", ".join(k + ": " + str(v) for k, v in self.attributes.items()) + " }"

    def __str__(self):
        return str(self.value)




import sys

import mlflow

from pypads import logger
from pypads.util import dict_merge


def cache_merge(*dicts):
    merged = {}
    for d in dicts:
        if isinstance(d, dict):
            for key, value in d.items():
                if isinstance(value, dict):
                    node = merged.setdefault(key, {})
                    merged[key] = dict_merge(node, value)
                elif isinstance(value, list):
                    node = merged.setdefault(key, [])
                    merged[key] = node.extend(value)
                elif isinstance(value, set):
                    s: set = merged.setdefault(key, set())
                    for v in value:
                        if v in s:
                            merged = dict_merge(v, s.pop(v))
                            s.add(merged)
                        else:
                            s.add(v)
                elif isinstance(value, Cache):
                    node = merged.setdefault(key, Cache())
                    merged[key] = value.merge(node)
                else:
                    merged[key] = value
    return merged


class Cache:
    def __init__(self):
        self._cache = {}

    @property
    def cache(self):
        return self._cache

    def merge(self, other):
        self._cache = cache_merge(self.cache, other.cache)
        return self

    def add(self, key, value):
        if key in self.cache:
            if isinstance(value, dict):
                self.cache.get(key).update(value)
            else:
                self.cache[key] = value
        else:
            self.cache.update({key: value})

    def pop(self, key, default=None):
        if key in self.cache:
            return self.cache.pop(key, default)
        return None

    def get(self, item, default=None):
        return self._cache.get(item,default)

    def items(self):
        return self._cache.items()

    def exists(self, key):
        return key in self._cache

    def clear(self):
        self._cache = {}

    def __str__(self):
        out = str(super(Cache, self).__str__()) + "["
        out += ",".join([str(k) + ": " + str(i) for k, i in self._cache.items()])
        out += "]"
        return out

    # def __getstate__(self):
    #     """
    #     Overwrite standard pickling by excluding the functions
    #     :return:
    #     """
    #     # can't pickle functions - use cloudpickle here?
    #     state = self.__dict__.copy()
    #     if "cache_cleanup" in self._cache:
    #         del state["_cache"]
    #     return state
    #
    # def __setstate__(self, state):
    #     self.__dict__.update(state)
    #     if hasattr(self, "_cache") or self._cache is None:
    #         self._cache = {}


class PypadsRunCache(Cache):
    def __init__(self, run_id):
        super().__init__()
        self._run_id = run_id or mlflow.active_run().info.run_id
        if not self._run_id:
            raise ValueError("No active run for run cache found.")

    @property
    def run_id(self):
        return self._run_id

    def register_cleanup_fn(self):
        from pypads.pypads import get_current_pads
        pads = get_current_pads()

        def cleanup_cache(*args, run_id=self.run_id, **kwargs):
            from pypads.pypads import get_current_pads
            pads = get_current_pads()
            pads.cache.run_delete(run_id)
            logger.debug("Cleared run cache after run " + run_id)

        pads.api.register_post_fn("cache_cleanup", cleanup_cache, order=sys.maxsize)


class PypadsCache(Cache):

    def __init__(self):
        super().__init__()
        self._run_caches = {}

    def merge(self, other):
        super().merge(other)
        self._run_caches = cache_merge(self.run_caches, other.run_caches)

    @property
    def run_caches(self):
        return self._run_caches

    def run_cache(self, run_id=None):
        run = self.run_init(run_id)
        return self._run_caches.get(run)

    def _get_run(self, run_id=None):
        if run_id is None:
            run_id = mlflow.active_run().info.run_id
        return run_id

    def run_init(self, run_id=None):
        if run_id is None:
            run_id = mlflow.active_run().info.run_id
        if not run_id:
            raise ValueError("No run is active. Couldn't init run cache.")
        if run_id not in self._run_caches:
            run_cache = PypadsRunCache(run_id)
            self._run_caches[run_id] = run_cache
            run_cache.register_cleanup_fn()
        return run_id

    def run_add(self, key, value, run_id=None):
        run_id = self.run_init(run_id)
        self._run_caches[run_id].add(key, value)

    def run_pop(self, key, run_id=None, default=None):
        run_id = self.run_init(run_id)
        return self._run_caches[run_id].pop(key, default=default)

    def run_remove(self, key, run_id=None):
        run_id = self.run_init(run_id)
        del self._run_caches[run_id][key]

    def run_get(self, key, run_id=None):
        run_id = self.run_init(run_id)
        return self._run_caches[run_id].get(key)

    def run_exists(self, *keys, run_id=None):
        run_id = self.run_init(run_id)
        return all([self._run_caches[run_id].exists(key) for key in keys])

    def run_clear(self, run_id=None):
        run_id = self.run_init(run_id)
        self._run_caches[run_id].clear()

    def run_delete(self, run_id=None):
        run_id = self.run_init(run_id)
        del self._run_caches[run_id]



import inspect

from pypads import logger
from pypads.autolog.hook import make_hook_applicable_filter
from pypads.autolog.wrapping.base_wrapper import BaseWrapper


class ModuleWrapper(BaseWrapper):

    def __init__(self, pypads):
        super().__init__(pypads)
        self._punched_module_names = set()

    @property
    def punched_module_names(self):
        return self._punched_module_names

    def wrap(self, module, context, mapping):
        """
        Function to wrap modules with pypads functionality
        :param module:
        :param mapping:
        :return:
        """
        if module.__name__ not in self.punched_module_names or not context.has_wrap_meta(mapping, module):
            self._pypads.wrap_manager.module_wrapper.punched_module_names.add(module.__name__)
            context.store_wrap_meta(mapping, module)

            if not context.has_original(module):
                context.store_original(module)

            # Get default hooks
            if not mapping.hooks:
                mapping.hooks = mapping.in_collection.get_default_module_hooks()

            # Try to wrap every attr of the module
            if mapping.hooks:
                for hook in mapping.hooks:
                    # Only get entries defined directly in this module
                    # https://stackoverflow.com/questions/22578509/python-get-only-classes-defined-in-imported-module-with-dir
                    for name in list(filter(make_hook_applicable_filter(hook, module, mapping),
                                            [m[0] for m in inspect.getmembers(module) if
                                             hasattr(m[1], "__module__") and m[1].__module__ == module.__name__])):
                        attr = getattr(module, name)

                        # Don't track imported modules
                        if not inspect.ismodule(attr):
                            self._pypads.wrap_manager.wrap(attr, module, mapping)

            # Add found classes to the tracking
            # for name in find_applicable_hooks(ctx, mapping):
            #     algorithm_mapping = AlgorithmMapping(mapping.reference + "." + name, mapping.library,
            #                                          mapping.algorithm,
            #                                          mapping.file, mapping.hooks)
            #     algorithm_mapping.in_collection = mapping.in_collection
            #     from pypads.autolog.wrapping.wrapping import _add_found_class
            #     _add_found_class(algorithm_mapping)
        else:
            logger.debug("Module " + str(module) + " already duck-punched.")
        return module
